import builtins
import google.protobuf.message
import typing

class ConsumerOptions(google.protobuf.message.Message):

    def __init__(self, *, fading: typing.Optional[builtins.bool]=None) -> None:...

    @builtins.property
    def fading(self) -> builtins.bool:...

    def HasField(self, field_name: typing.Literal['fading']) -> builtins.bool:...

    def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...


from __future__ import annotations

import asyncio
import typing as t

if t.TYPE_CHECKING:
    from concurrent.futures import Executor

    from brokrpc.rpc.abc import HandlerSerializer
    from brokrpc.rpc.model import BinaryResponse, Request
    from brokrpc.rpc.storage import WaiterStorage

from brokrpc.abc import Consumer, Publisher
from brokrpc.message import AppMessage, BinaryMessage, Message
from brokrpc.model import ConsumerAck, ConsumerReject, ConsumerResult, PublisherResult


class SyncFuncHandler[U, V](Consumer[BinaryMessage, ConsumerResult]):
    def __init__(
        self,
        func: t.Callable[[Request[U]], V],
        serializer: HandlerSerializer[U, V],
        replier: Publisher[BinaryMessage, PublisherResult],
        executor: Executor | None = None,
    ) -> None:
        self.__func = func
        self.__serializer = serializer
        self.__replier = replier
        self.__executor = executor

    async def consume(self, request_message: BinaryMessage) -> ConsumerResult:
        handle_result, response_message = await asyncio.get_running_loop().run_in_executor(
            self.__executor,
            self.__handle,
            request_message,
        )

        if isinstance(handle_result, ConsumerAck):
            await self.__replier.publish(response_message)

        return handle_result

    def __handle(self, request_message: BinaryMessage) -> tuple[ConsumerResult, BinaryMessage]:
        response_factory = get_response_factory(request_message, self.__serializer)
        if response_factory is None:
            return reject_request_no_reply(request_message), request_message

        request_payload = self.__serializer.load_unary_request(request_message)
        response_payload = self.__func(request_payload)
        response_message = response_factory(response_payload)

        return ConsumerAck(), response_message


class AsyncFuncHandler[U, V](Consumer[BinaryMessage, ConsumerResult]):
    def __init__(
        self,
        func: t.Callable[[Message[U]], t.Awaitable[V]],
        serializer: HandlerSerializer[U, V],
        replier: Publisher[BinaryMessage, PublisherResult],
    ) -> None:
        self.__func = func
        self.__serializer = serializer
        self.__replier = replier

    async def consume(self, request_message: BinaryMessage) -> ConsumerResult:
        response_factory = get_response_factory(request_message, self.__serializer)
        if response_factory is None:
            return reject_request_no_reply(request_message)

        request_payload = self.__serializer.load_unary_request(request_message)
        response_payload = await self.__func(request_payload)
        response_message = response_factory(response_payload)
        await self.__replier.publish(response_message)

        return ConsumerAck()


class HandlerResponseConsumer[T](Consumer[BinaryMessage, ConsumerResult]):
    def __init__(self, serializer: t.Callable[[BinaryMessage], T], storage: WaiterStorage[T]) -> None:
        self.__serializer = serializer
        self.__storage = storage

    async def consume(self, response_message: BinaryMessage) -> ConsumerResult:
        if response_message.correlation_id is None:
            return ConsumerReject(reason=f"can't get correlation id from response message: {response_message}")

        result: T | Exception
        try:
            result = self.__serializer(response_message)

        # NOTE: caught exception is propagated via storage waiter
        except Exception as err:  # noqa: BLE001
            result = err

        self.__storage.set_waiter(response_message.correlation_id, result)

        return ConsumerAck()


def get_response_factory[U, V](
    request: Request[U],
    serializer: HandlerSerializer[U, V],
) -> t.Callable[[V], BinaryResponse] | None:
    if (reply_to := request.reply_to) is not None and (correlation_id := request.correlation_id) is not None:

        def create(response_payload: V) -> BinaryResponse:
            reply = AppMessage(
                body=response_payload,
                routing_key=reply_to,
                exchange=request.exchange,
                correlation_id=correlation_id,
                timeout=request.timeout,
            )
            response_message = serializer.dump_unary_response(reply)

            # NOTE: `response_message` var is for debugger.
            return response_message  # noqa: RET504

        return create

    return None


def reject_request_no_reply(request_message: Message[object]) -> ConsumerResult:
    return ConsumerReject(reason=f"can't get reply options from request message: {request_message}")



from __future__ import annotations

import abc
import typing as t

if t.TYPE_CHECKING:
    from datetime import datetime, timedelta

from brokrpc.debug import is_debug_enabled
from brokrpc.stringify import to_str_obj


class Message[T](metaclass=abc.ABCMeta):
    if not is_debug_enabled():

        def __str__(self) -> str:
            return to_str_obj(self, type_vars=[type(self.body)], id=self.message_id)

    else:

        def __str__(self) -> str:
            return repr(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.body == other.body
            and self.routing_key == other.routing_key
            and self.exchange == other.exchange
            and self.content_type == other.content_type
            and self.content_encoding == other.content_encoding
            and self.headers == other.headers
            and self.delivery_mode == other.delivery_mode
            and self.priority == other.priority
            and self.correlation_id == other.correlation_id
            and self.reply_to == other.reply_to
            and self.timeout == other.timeout
            and self.message_id == other.message_id
            and self.timestamp == other.timestamp
            and self.message_type == other.message_type
            and self.user_id == other.user_id
            and self.app_id == other.app_id
        )

    @property
    @abc.abstractmethod
    def body(self) -> T:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def routing_key(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def exchange(self) -> str | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def content_type(self) -> str | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def content_encoding(self) -> str | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def headers(self) -> t.Mapping[str, str] | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def delivery_mode(self) -> int | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def priority(self) -> int | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def correlation_id(self) -> str | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def reply_to(self) -> str | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def timeout(self) -> timedelta | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def message_id(self) -> str | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def timestamp(self) -> datetime | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def message_type(self) -> str | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def user_id(self) -> str | None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def app_id(self) -> str | None:
        raise NotImplementedError


type BinaryMessage = Message[bytes]


class AppMessage[T](Message[T]):
    __slots__ = (
        "__body",
        "__routing_key",
        "__exchange",
        "__content_type",
        "__content_encoding",
        "__headers",
        "__delivery_mode",
        "__priority",
        "__correlation_id",
        "__reply_to",
        "__timeout",
        "__message_id",
        "__timestamp",
        "__message_type",
        "__user_id",
        "__app_id",
    )

    # NOTE: message constructor has a lot of options to set up a structure (dataclass)
    def __init__(  # noqa: PLR0913
        self,
        *,
        body: T,
        routing_key: str,
        exchange: str | None = None,
        content_type: str | None = None,
        content_encoding: str | None = None,
        headers: t.Mapping[str, str] | None = None,
        delivery_mode: int | None = None,
        priority: int | None = None,
        correlation_id: str | None = None,
        reply_to: str | None = None,
        timeout: timedelta | None = None,
        message_id: str | None = None,
        timestamp: datetime | None = None,
        message_type: str | None = None,
        user_id: str | None = None,
        app_id: str | None = None,
    ) -> None:
        self.__body = body
        self.__routing_key = routing_key
        self.__exchange = exchange
        self.__content_type = content_type
        self.__content_encoding = content_encoding
        self.__headers = headers
        self.__delivery_mode = delivery_mode
        self.__priority = priority
        self.__correlation_id = correlation_id
        self.__reply_to = reply_to
        self.__timeout = timeout
        self.__message_id = message_id
        self.__timestamp = timestamp
        self.__message_type = message_type
        self.__user_id = user_id
        self.__app_id = app_id

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"body={self.__body!r}, "
            f"routing_key={self.__routing_key!r}, "
            f"exchange={self.__exchange!r}, "
            f"content_type={self.__content_type!r}, "
            f"content_encoding={self.__content_encoding!r}, "
            f"headers={self.__headers!r}, "
            f"delivery_mode={self.__delivery_mode!r}, "
            f"priority={self.__priority!r}, "
            f"correlation_id={self.__correlation_id!r}, "
            f"reply_to={self.__reply_to!r}, "
            f"timeout={self.__timeout!r}, "
            f"message_id={self.__message_id!r}, "
            f"timestamp={self.__timestamp!r}, "
            f"message_type={self.__message_type!r}, "
            f"user_id={self.__user_id!r}, "
            f"app_id={self.__app_id!r}"
            ")"
        )

    @property
    def body(self) -> T:
        return self.__body

    @property
    def routing_key(self) -> str:
        return self.__routing_key

    @property
    def exchange(self) -> str | None:
        return self.__exchange

    @property
    def content_type(self) -> str | None:
        return self.__content_type

    @property
    def content_encoding(self) -> str | None:
        return self.__content_encoding

    @property
    def headers(self) -> t.Mapping[str, str] | None:
        return self.__headers

    @property
    def delivery_mode(self) -> int | None:
        return self.__delivery_mode

    @property
    def priority(self) -> int | None:
        return self.__priority

    @property
    def correlation_id(self) -> str | None:
        return self.__correlation_id

    @property
    def reply_to(self) -> str | None:
        return self.__reply_to

    @property
    def timeout(self) -> timedelta | None:
        return self.__timeout

    @property
    def message_id(self) -> str | None:
        return self.__message_id

    @property
    def timestamp(self) -> datetime | None:
        return self.__timestamp

    @property
    def message_type(self) -> str | None:
        return self.__message_type

    @property
    def user_id(self) -> str | None:
        return self.__user_id

    @property
    def app_id(self) -> str | None:
        return self.__app_id


@t.final
class PackedMessage[T](Message[T]):
    __slots__ = (
        "__original",
        "__body",
        "__content_type",
        "__content_encoding",
        "__headers",
        "__message_type",
    )

    # NOTE: message constructor has a lot of options to set up a structure (dataclass)
    def __init__(  # noqa: PLR0913
        self,
        *,
        body: T,
        content_type: str | None = None,
        content_encoding: str | None = None,
        headers: t.Mapping[str, str] | None = None,
        message_type: str | None = None,
        original: Message[object],
    ) -> None:
        self.__body = body
        self.__content_type = content_type if content_type is not None else original.content_type
        self.__content_encoding = content_encoding if content_encoding is not None else original.content_encoding
        self.__headers = headers if headers is not None else original.headers
        self.__message_type = message_type if message_type is not None else original.message_type
        self.__original = original

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"body={self.__body!r}, "
            f"content_type={self.__content_type!r}, "
            f"content_encoding={self.__content_encoding!r}, "
            f"headers={self.__headers!r}, "
            f"message_type={self.__message_type!r}, "
            f"original={self.__original!r}"
            ")"
        )

    @property
    def body(self) -> T:
        return self.__body

    @property
    def routing_key(self) -> str:
        return self.__original.routing_key

    @property
    def exchange(self) -> str | None:
        return self.__original.exchange

    @property
    def content_type(self) -> str | None:
        return self.__content_type

    @property
    def content_encoding(self) -> str | None:
        return self.__content_encoding

    @property
    def headers(self) -> t.Mapping[str, str] | None:
        return self.__headers

    @property
    def delivery_mode(self) -> int | None:
        return self.__original.delivery_mode

    @property
    def priority(self) -> int | None:
        return self.__original.priority

    @property
    def correlation_id(self) -> str | None:
        return self.__original.correlation_id

    @property
    def reply_to(self) -> str | None:
        return self.__original.reply_to

    @property
    def timeout(self) -> timedelta | None:
        return self.__original.timeout

    @property
    def message_id(self) -> str | None:
        return self.__original.message_id

    @property
    def timestamp(self) -> datetime | None:
        return self.__original.timestamp

    @property
    def message_type(self) -> str | None:
        return self.__message_type

    @property
    def user_id(self) -> str | None:
        return self.__original.user_id

    @property
    def app_id(self) -> str | None:
        return self.__original.app_id


@t.final
class UnpackedMessage[T](Message[T]):
    __slots__ = (
        "__original",
        "__body",
    )

    def __init__(self, original: Message[object], body: T) -> None:
        self.__original = original
        self.__body = body

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(original={self.__original!r}, body={self.__body!r})"

    @property
    def original(self) -> Message[object]:
        return self.__original

    @property
    def body(self) -> T:
        return self.__body

    @property
    def routing_key(self) -> str:
        return self.__original.routing_key

    @property
    def exchange(self) -> str | None:
        return self.__original.exchange

    @property
    def content_type(self) -> str | None:
        return self.__original.content_type

    @property
    def content_encoding(self) -> str | None:
        return self.__original.content_encoding

    @property
    def headers(self) -> t.Mapping[str, str] | None:
        return self.__original.headers

    @property
    def delivery_mode(self) -> int | None:
        return self.__original.delivery_mode

    @property
    def priority(self) -> int | None:
        return self.__original.priority

    @property
    def correlation_id(self) -> str | None:
        return self.__original.correlation_id

    @property
    def reply_to(self) -> str | None:
        return self.__original.reply_to

    @property
    def timeout(self) -> timedelta | None:
        return self.__original.timeout

    @property
    def message_id(self) -> str | None:
        return self.__original.message_id

    @property
    def timestamp(self) -> datetime | None:
        return self.__original.timestamp

    @property
    def message_type(self) -> str | None:
        return self.__original.message_type

    @property
    def user_id(self) -> str | None:
        return self.__original.user_id

    @property
    def app_id(self) -> str | None:
        return self.__original.app_id



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: brokrpc/spec/v1/consumer.proto
# Protobuf Python Version: 5.27.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    1,
    '',
    'brokrpc/spec/v1/consumer.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1e\x62rokrpc/spec/v1/consumer.proto\x12\x0f\x62rokrpc.spec.v1\x1a google/protobuf/descriptor.proto\"9\n\x0f\x43onsumerOptions\x12\x1b\n\x06\x66\x61\x64ing\x18\x01 \x01(\x08H\x00R\x06\x66\x61\x64ing\x88\x01\x01\x42\t\n\x07_fading:`\n\x08\x63onsumer\x12\x1e.google.protobuf.MethodOptions\x18\xa9\x46 \x01(\x0b\x32 .brokrpc.spec.v1.ConsumerOptionsR\x08\x63onsumer\x88\x01\x01\x42\x82\x01\n\x13\x63om.brokrpc.spec.v1B\rConsumerProtoP\x01\xa2\x02\x03\x42SX\xaa\x02\x0f\x42rokrpc.Spec.V1\xca\x02\x0f\x42rokrpc\\Spec\\V1\xe2\x02\x1b\x42rokrpc\\Spec\\V1\\GPBMetadata\xea\x02\x11\x42rokrpc::Spec::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'brokrpc.spec.v1.consumer_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\023com.brokrpc.spec.v1B\rConsumerProtoP\001\242\002\003BSX\252\002\017Brokrpc.Spec.V1\312\002\017Brokrpc\\Spec\\V1\342\002\033Brokrpc\\Spec\\V1\\GPBMetadata\352\002\021Brokrpc::Spec::V1'
  _globals['_CONSUMEROPTIONS']._serialized_start=85
  _globals['_CONSUMEROPTIONS']._serialized_end=142
# @@protoc_insertion_point(module_scope)



import importlib
import typing as t
from functools import cached_property


@t.final
class Loader[T]:
    def load_class(self, entrypoint: str) -> type[T]:
        klass = self.__load(entrypoint)
        if not self.__check_class(klass):
            details = "invalid class type"
            raise TypeError(details, klass)

        return klass

    def load_instance(self, entrypoint: str) -> T:
        factory = self.__load(entrypoint)
        if not callable(factory):
            details = "invalid factory type"
            raise TypeError(details, factory)

        obj = factory()
        if not self.__check_instance(obj):
            details = "invalid instance type"
            raise TypeError(details, obj)

        return obj

    def load_constant(self, entrypoint: str) -> T:
        obj = self.__load(entrypoint)
        if not self.__check_instance(obj):
            details = "invalid instance type"
            raise TypeError(details, obj)

        return obj

    @cached_property
    def __type(self) -> type[T]:
        # NOTE: this hack with `__orig_class__` allows to pass interfaces to `T` (classes that derives `abc.ABC`
        # directly and has `abc.abstractmethod` decorators). See: https://github.com/python/mypy/issues/4717
        type_args = t.get_args(getattr(self, "__orig_class__", None))
        assert len(type_args) == 1

        type_ = t.get_origin(type_args[0]) or type_args[0]
        assert isinstance(type_, type)

        return type_

    def __load(self, entrypoint: str) -> object:
        match entrypoint.split(":", maxsplit=1):
            case [module, attr]:
                try:
                    loaded_module = importlib.import_module(module)

                except ImportError as err:
                    details = "invalid entrypoint"
                    raise ValueError(details, entrypoint) from err

                try:
                    obj = getattr(loaded_module, attr)

                except AttributeError as err:
                    details = "invalid entrypoint"
                    raise ValueError(details, entrypoint) from err

                return obj

            case _:
                details = "invalid entrypoint"
                raise ValueError(details, entrypoint)

    def __check_class(self, obj: object) -> t.TypeGuard[type[T]]:
        return isinstance(obj, type) and issubclass(obj, self.__type)

    def __check_instance(self, obj: object) -> t.TypeGuard[T]:
        return isinstance(obj, self.__type)



import abc

from brokrpc.rpc.model import BinaryRequest, BinaryResponse, Request, Response


class UnaryUnaryHandler[U, V](metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def handle(self, request: U) -> V:
        raise NotImplementedError


class Caller[U, V](metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def invoke(self, request: U) -> Response[V]:
        raise NotImplementedError


class HandlerSerializer[U, V](metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def load_unary_request(self, request: BinaryRequest) -> Request[U]:
        raise NotImplementedError

    @abc.abstractmethod
    def dump_unary_response(self, response: Response[V]) -> BinaryResponse:
        raise NotImplementedError


class CallerSerializer[U, V](metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def dump_unary_request(self, request: Request[U]) -> BinaryRequest:
        raise NotImplementedError

    @abc.abstractmethod
    def load_unary_response(self, response: BinaryResponse) -> Response[V]:
        raise NotImplementedError


class RPCSerializer[U, V](HandlerSerializer[U, V], CallerSerializer[U, V], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def dump_unary_request(self, request: Request[U]) -> BinaryRequest:
        raise NotImplementedError

    @abc.abstractmethod
    def load_unary_request(self, request: BinaryRequest) -> Request[U]:
        raise NotImplementedError

    @abc.abstractmethod
    def dump_unary_response(self, response: Response[V]) -> BinaryResponse:
        raise NotImplementedError

    @abc.abstractmethod
    def load_unary_response(self, response: BinaryResponse) -> Response[V]:
        raise NotImplementedError

    # @abc.abstractmethod
    # def dump_stream_request(
    #     self,
    #     request: t.Sequence[U],
    #     options: MessageOptions,
    # ) -> t.Sequence[StreamRequest]:
    #     raise NotImplementedError
    #
    # @abc.abstractmethod
    # def load_stream_request(self, request: t.Sequence[StreamRequest]) -> t.Sequence[U]:
    #     raise NotImplementedError
    #
    # @abc.abstractmethod
    # def dump_stream_response(
    #     self,
    #     request: StreamRequest,
    #     response: t.Sequence[V | Exception],
    #     options: MessageOptions,
    # ) -> t.Sequence[StreamResponse]:
    #     raise NotImplementedError
    #
    # @abc.abstractmethod
    # def load_stream_response(self, response: t.Sequence[StreamResponse]) -> t.Sequence[V | Exception]:
    #     raise NotImplementedError



import asyncio
import sys
import typing as t
from argparse import ArgumentParser
from functools import partial
from signal import Signals
from uuid import uuid4

from aiofiles import stderr_bytes, stdin_bytes, stdout_bytes
from aiofiles.threadpool.binary import AsyncIndirectBufferedIOBase
from yarl import URL

from brokrpc.abc import Serializer
from brokrpc.broker import Broker
from brokrpc.entrypoint import Loader
from brokrpc.message import AppMessage, BinaryMessage, Message
from brokrpc.middleware import AbortBadMessageMiddleware
from brokrpc.options import BindingOptions, ExchangeOptions, PublisherOptions, QueueOptions
from brokrpc.serializer.ident import IdentSerializer


class CLIOptions:
    url: URL
    exchange: ExchangeOptions | None
    routing_key: str
    kind: t.Literal["consumer", "publisher"]
    serializer: Serializer[Message[object], BinaryMessage]
    output_mode: t.Literal["wide", "body"]
    quiet: bool
    input: AsyncIndirectBufferedIOBase
    output: AsyncIndirectBufferedIOBase
    err: AsyncIndirectBufferedIOBase
    done: asyncio.Event

    def __str__(self) -> str:
        data = ", ".join(f"{k}={v!r}" for k, v in self.__dict__.items())
        return f"{self.__class__.__name__}({data})"

    __repr__ = __str__


async def run(options: CLIOptions) -> int:
    async with (
        Broker(
            options.url,
            default_publisher_middlewares=[
                AbortBadMessageMiddleware(),
            ],
            default_consumer_middlewares=[
                AbortBadMessageMiddleware(),
            ],
        ) as broker,
    ):
        match options.kind:
            case "consumer":
                return await run_consumer(broker, options)

            case "publisher":
                return await run_publisher(broker, options)

            case _:
                t.assert_never(options.kind)


async def run_consumer(broker: Broker, options: CLIOptions) -> int:
    async with broker.consumer(
        partial(consume_message, options.output)
        if options.output_mode == "wide"
        else partial(consume_message_body, options.output),
        BindingOptions(
            exchange=options.exchange,
            binding_keys=[options.routing_key],
            queue=QueueOptions(
                name=f"{__name__}.{uuid4().hex}",
                exclusive=True,
                auto_delete=True,
                prefetch_count=1,
            ),
        ),
        serializer=options.serializer,
    ):
        await options.done.wait()

        return 0


async def run_publisher(broker: Broker, options: CLIOptions) -> int:
    async with broker.publisher(
        PublisherOptions(
            name=options.exchange.name,
            prefetch_count=1,
        )
        if options.exchange is not None
        else PublisherOptions(prefetch_count=1),
        serializer=options.serializer,
    ) as publisher:
        async for body in options.input:
            message = AppMessage(body=body, routing_key=options.routing_key)
            result = await publisher.publish(message)

            match result:
                case True | None:
                    if not options.quiet:
                        await options.err.write(b"sent\n")
                        await options.err.flush()

                case False:
                    if not options.quiet:
                        await options.err.write(f"failed to send: {message!r}\n".encode())
                        await options.err.flush()

                case _:
                    t.assert_never(result)

        return 0


async def consume_message(out: AsyncIndirectBufferedIOBase, message: Message[object]) -> None:
    await out.write(f"{message!r}\n".encode())
    await out.flush()


async def consume_message_body(out: AsyncIndirectBufferedIOBase, message: Message[bytes]) -> None:
    await out.write(message.body + b"\n")
    await out.flush()


def build_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="""Exchange messages with broker.""",
    )
    parser.add_argument(
        "url",
        type=URL,
        help="broker url to connect to",
    )
    parser.add_argument(
        "routing_key",
        type=str,
        help="routing key to use for message publishing / consumption",
    )
    parser.add_argument(
        "-e",
        "--exchange",
        type=parse_exchange_name,
        help="exchange name to connect to",
        default=None,
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-c",
        "--consumer",
        dest="kind",
        action="store_const",
        const="consumer",
        help="consumer mode. Print received messages to stdout.",
    )
    group.add_argument(
        "-p",
        "--publisher",
        dest="kind",
        action="store_const",
        const="publisher",
        help="publisher mode. Send messages from stdin.",
    )

    parser.add_argument(
        "-s",
        "--serializer",
        type=parse_serializer,
        help="customize serializer. Available options: json, protobuf:{entrypoint to protobuf message class}, "
        "{entrypoint to serializer class}. Default: none (pass binary data to body as is).",
        default=IdentSerializer(),
    )

    parser.add_argument(
        "-o",
        "--output",
        type=str,
        dest="output_mode",
        choices=["wide", "body"],
        help="customize output mode. Available options: wide, body. Default: body",
        default="body",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_const",
        const=True,
        help="don't print to stderr",
        default=False,
    )

    return parser


def parse_exchange_name(value: str) -> ExchangeOptions:
    return ExchangeOptions(name=value)


def parse_serializer(value: str | None) -> Serializer[t.Any, BinaryMessage]:
    if value is None:
        return IdentSerializer()

    match value.split(":", maxsplit=1):
        case ["raw"]:
            return IdentSerializer()

        case ["json"]:
            from brokrpc.serializer.json import JSONSerializer

            return JSONSerializer()

        case ["protobuf", msg_type]:
            from google.protobuf.message import Message as ProtobufMessage

            from brokrpc.serializer.protobuf import JSONProtobufSerializer

            return JSONProtobufSerializer(Loader[ProtobufMessage]().load_class(msg_type))

        case _:
            return Loader[Serializer[t.Any, BinaryMessage]]().load_instance(value)


def main() -> None:
    options = build_parser().parse_args(namespace=CLIOptions())

    options.input = stdin_bytes
    options.output = stdout_bytes
    options.err = stderr_bytes
    done = options.done = asyncio.Event()

    for sig in (Signals.SIGINT, Signals.SIGTERM):
        asyncio.get_event_loop().add_signal_handler(sig, done.set)

    sys.exit(asyncio.run(run(options)))


if __name__ == "__main__":
    main()



from __future__ import annotations

import typing as t

from brokrpc.abc import Publisher
from brokrpc.stringify import to_str_obj


class EncodingPublisher[A, V, B](Publisher[B, V]):
    def __init__(self, inner: Publisher[A, V], encoder: t.Callable[[B], A]) -> None:
        self.__inner = inner
        self.__encoder = encoder

    def __str__(self) -> str:
        return to_str_obj(self, inner=self.__inner)

    async def publish(self, message: B) -> V:
        encoded_message = self.__encoder(message)
        result = await self.__inner.publish(encoded_message)

        # NOTE: `result` var is for debugger.
        return result  # noqa: RET504


class DecodingPublisher[U, A, B](Publisher[U, B]):
    def __init__(self, inner: Publisher[U, A], decoder: t.Callable[[A], B]) -> None:
        self.__inner = inner
        self.__decoder = decoder

    def __str__(self) -> str:
        return to_str_obj(self, inner=self.__inner)

    async def publish(self, message: U) -> B:
        result = await self.__inner.publish(message)
        decoded_result = self.__decoder(result)

        # NOTE: `decoded_result` var is for debugger.
        return decoded_result  # noqa: RET504



from __future__ import annotations

import asyncio
import typing as t

if t.TYPE_CHECKING:
    from concurrent.futures import Executor

from brokrpc.abc import Consumer
from brokrpc.stringify import to_str_obj


class DecodingConsumer[A, V, B](Consumer[A, V]):
    def __init__(self, inner: Consumer[B, V], decoder: t.Callable[[A], B]) -> None:
        self.__inner = inner
        self.__decoder = decoder

    def __str__(self) -> str:
        return to_str_obj(self, inner=self.__inner)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(inner={self.__inner!r}, decoder={self.__decoder!r})"

    async def consume(self, message: A) -> V:
        decoded_message = self.__decoder(message)
        result = await self.__inner.consume(decoded_message)

        # NOTE: `result` var is for debugger.
        return result  # noqa: RET504


class EncodingConsumer[U, A, B](Consumer[U, A]):
    def __init__(self, inner: Consumer[U, B], encode: t.Callable[[B], A]) -> None:
        self.__inner = inner
        self.__encode = encode

    def __str__(self) -> str:
        return to_str_obj(self, inner=self.__inner)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(inner={self.__inner!r}, encode={self.__encode!r})"

    async def consume(self, message: U) -> A:
        result = await self.__inner.consume(message)
        encoded_result = self.__encode(result)

        # NOTE: `encoded_result` var is for debugger.
        return encoded_result  # noqa: RET504


class SyncFuncConsumer[U, V](Consumer[U, V]):
    def __init__(self, func: t.Callable[[U], V], executor: Executor | None = None) -> None:
        self.__func = func
        self.__executor = executor

    def __str__(self) -> str:
        return to_str_obj(self, inner=self.__func)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(func={self.__func!r}, executor={self.__executor!r})"

    async def consume(self, message: U) -> V:
        result = await asyncio.get_running_loop().run_in_executor(self.__executor, self.__func, message)

        # NOTE: `result` var is for debugger.
        return result  # noqa: RET504


class AsyncFuncConsumer[U, V](Consumer[U, V]):
    def __init__(self, func: t.Callable[[U], t.Awaitable[V]]) -> None:
        self.__func = func

    def __str__(self) -> str:
        return to_str_obj(self, inner=self.__func)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(func={self.__func!r})"

    async def consume(self, message: U) -> V:
        result = await self.__func(message)

        # NOTE: `result` var is for debugger.
        return result  # noqa: RET504



from brokrpc.rpc.abc import CallerSerializer, HandlerSerializer
from brokrpc.rpc.model import BinaryRequest, BinaryResponse


class RawRPCSerializer(CallerSerializer[bytes, bytes], HandlerSerializer[bytes, bytes]):
    def dump_unary_request(self, request: BinaryRequest) -> BinaryRequest:
        return request

    def load_unary_request(self, request: BinaryRequest) -> BinaryRequest:
        return request

    def dump_unary_response(self, response: BinaryResponse) -> BinaryResponse:
        return response

    def load_unary_response(self, response: BinaryResponse) -> BinaryResponse:
        return response



from __future__ import annotations

import asyncio
import typing as t

if t.TYPE_CHECKING:
    from datetime import timedelta

    from brokrpc.abc import BinaryPublisher
    from brokrpc.options import ExchangeOptions
    from brokrpc.rpc.model import BinaryRequest, Request, Response
    from brokrpc.rpc.storage import WaiterStorage


from brokrpc.message import AppMessage
from brokrpc.rpc.abc import Caller
from brokrpc.stringify import to_str_obj


class RequestCaller[U, V](Caller[U, V]):
    # TODO: shorter arguments
    def __init__(  # noqa: PLR0913
        self,
        requester: BinaryPublisher,
        routing_key: str,
        serializer: t.Callable[[Request[U]], BinaryRequest],
        reply_to: str,
        storage: WaiterStorage[Response[V]],
        exchange: ExchangeOptions | None = None,
        timeout: timedelta | None = None,
    ) -> None:
        self.__requester = requester
        self.__routing_key = routing_key
        self.__serializer = serializer
        self.__reply_to = reply_to
        self.__storage = storage
        self.__exchange = exchange
        self.__timeout = timeout

    def __str__(self) -> str:
        return to_str_obj(self, requester=self.__requester, routing_key=self.__routing_key, reply_to=self.__reply_to)

    async def invoke(self, request_payload: U) -> Response[V]:
        waiter: asyncio.Future[Response[V]]

        with self.__storage.context() as (correlation_id, waiter):
            request_message = self.__serializer(
                AppMessage(
                    body=request_payload,
                    exchange=self.__exchange.name if self.__exchange is not None else None,
                    routing_key=self.__routing_key,
                    correlation_id=correlation_id,
                    reply_to=self.__reply_to,
                    timeout=self.__timeout,
                )
            )

            async with asyncio.timeout(
                request_message.timeout.total_seconds() if request_message.timeout is not None else None
            ):
                publish_result = await self.__requester.publish(request_message)
                if publish_result is False:
                    # TODO: raise appropriate exception
                    raise RuntimeError(...)

                response = await waiter

        return response



from __future__ import annotations

import typing as t
from dataclasses import asdict, dataclass, replace

if t.TYPE_CHECKING:
    from _typeshed import DataclassInstance
    from yarl import URL


from brokrpc.retry import RetryOptions


class DecoratorOptions:
    @classmethod
    def load(cls, obj: object) -> t.Self:
        return t.cast(t.Self, getattr(obj, cls._get_key(), cls()))

    def __call__[**U, V](self, func: t.Callable[U, V]) -> t.Callable[U, V]:
        setattr(func, self._get_key(), self)
        return func

    @classmethod
    def _get_key(cls) -> str:
        return f"{cls.__module__}.{cls.__name__}"


type ExchangeType = t.Literal["direct", "fanout", "topic", "headers"]


@dataclass(frozen=True, kw_only=True)
class BrokerOptions(RetryOptions):
    driver: str
    url: URL


@dataclass(frozen=True, kw_only=True)
class BrokerEntityOptions:
    mode: t.Literal["soft", "strict"] | None = None
    create_if_not_exist: bool | None = None
    name: str | None = None


@dataclass(frozen=True, kw_only=True)
class QOSOptions:
    prefetch_count: int | None = None

    def __post_init__(self) -> None:
        if self.prefetch_count is not None and self.prefetch_count < 0:
            details = "prefetch count must be greater than or equal to 0"
            raise ValueError(details, self)


@dataclass(frozen=True, kw_only=True)
class ExchangeOptions(BrokerEntityOptions, QOSOptions):
    type: ExchangeType | None = None
    durable: bool | None = None
    auto_delete: bool | None = None
    arguments: t.Mapping[str, object] | None = None


@dataclass(frozen=True, kw_only=True)
class QueueOptions(BrokerEntityOptions, QOSOptions):
    durable: bool | None = None
    exclusive: bool | None = None
    auto_delete: bool | None = None
    arguments: t.Mapping[str, object] | None = None


@dataclass(frozen=True, kw_only=True)
class BindingOptions:
    exchange: ExchangeOptions | None = None
    binding_keys: t.Sequence[str]
    queue: QueueOptions | None = None

    def __post_init__(self) -> None:
        if isinstance(self.binding_keys, str):
            details = (
                "a string (a sequence of chars) can't be used as a list of binding keys, it's a possible typo in "
                "the code (missed comma in tuple), try to provide a sequence of binding keys explicitly"
            )
            # NOTE: `str` is a valid type for `t.Sequence[str]`, but this value for binding_keys is not good.
            raise ValueError(details, list(self.binding_keys))  # noqa: TRY004

        if not self.binding_keys:
            details = "at least one binding key must be set"
            raise ValueError(details, self)


@dataclass(frozen=True, kw_only=True)
class ConsumerOptions(QueueOptions, RetryOptions, DecoratorOptions):
    retry_on_error: bool | type[Exception] | tuple[type[Exception], ...] | None = None


@dataclass(frozen=True, kw_only=True)
class PublisherOptions(ExchangeOptions):
    mandatory: bool | None = None


def merge_options[T: DataclassInstance](*options: T | None) -> T | None:
    options_iter = iter(options)

    result: T | None = None

    for result in options_iter:
        if result is not None:
            break

    if result is None:
        return None

    for opt in options_iter:
        if opt is None:
            continue

        result = replace(result, **{k: v for k, v in asdict(opt).items() if v is not None})

    return result



from __future__ import annotations

import asyncio
import typing as t

if t.TYPE_CHECKING:
    from datetime import timedelta

from dataclasses import dataclass
from itertools import count


@dataclass(frozen=True, kw_only=True)
class RetryOptions:
    retry_delay_mode: t.Literal["constant", "multiplier", "exponential"] | None = None
    retry_delay: timedelta | None = None
    retry_max_delay: timedelta | None = None
    retries_timeout: timedelta | None = None
    retries_limit: int | None = None

    def __post_init__(self) -> None:
        if self.retry_delay is not None and self.retry_delay.total_seconds() < 0.0:
            details = "retry delay must be greater than or equal to 0"
            raise ValueError(details, self)

        if self.retry_max_delay is not None and self.retry_max_delay.total_seconds() < 0.0:
            details = "retry max delay must be greater than or equal to 0"
            raise ValueError(details, self)

        if self.retries_timeout is not None and self.retries_timeout.total_seconds() <= 0.0:
            details = "retries timeout must be greater than 0"
            raise ValueError(details, self)

        if self.retries_limit is not None and self.retries_limit < 0:
            details = "retries limit must be greater than or equal to 0"
            raise ValueError(details, self)

        if (
            self.retry_max_delay is not None
            and self.retry_delay is not None
            and self.retry_delay > self.retry_max_delay
        ):
            details = "retry delay must be less than or equal to retry max delay"
            raise ValueError(details, self)


class Attempt:
    def __init__(self, n: int, errors: list[Exception]) -> None:
        self.__n = n
        self.__errors = errors
        self.__ok: bool | None = None

    @property
    def ok(self) -> bool | None:
        return self.__ok

    @property
    def num(self) -> int:
        return self.__n

    @property
    def is_retry(self) -> bool:
        return self.__n > 0

    def set_failed(self, err: Exception | None) -> None:
        if self.__ok is not None:
            return

        self.__ok = False
        if err is not None:
            self.__errors.append(err)

    def set_succeeded(self) -> None:
        if self.__ok is not None:
            return

        self.__ok = True


class RetryerError(Exception):
    pass


class NoMoreAttemptsError(ExceptionGroup, RetryerError):
    pass


class BoundRetryer:
    def __init__(
        self,
        retryer: Retryer,
        retry_on_exceptions: tuple[type[Exception], ...],
        no_more_attempts_message: str,
    ) -> None:
        self.__retryer = retryer
        self.__retry_on_exceptions = retry_on_exceptions
        self.__no_more_attempts_message = no_more_attempts_message

    async def do[**U, V](self, func: t.Callable[U, t.Awaitable[V]], *args: U.args, **kwargs: U.kwargs) -> V:
        async for attempt in self.__retryer.iterate(self.__no_more_attempts_message):
            try:
                result = await func(*args, **kwargs)

            except self.__retry_on_exceptions as err:
                attempt.set_failed(err)

            else:
                attempt.set_succeeded()
                return result

        # NOTE: it's always at least one attempt.
        raise RuntimeError


class Retryer:
    def __init__(self, options: RetryOptions) -> None:
        self.__options = options

    async def iterate(self, no_more_attempts_message: str) -> t.AsyncIterator[Attempt]:
        errors: list[Exception] = []

        timeout = self.__options.retries_timeout.total_seconds() if self.__options.retries_timeout is not None else None
        try:
            async with asyncio.timeout(timeout):
                async for attempt in self.__gen_attempts(errors):
                    yield attempt

                    if attempt.ok:
                        return

        except TimeoutError as err:
            errors.append(err)

        raise NoMoreAttemptsError(no_more_attempts_message, errors)

    def bind(
        self,
        retry_on_exceptions: tuple[type[Exception], ...],
        no_more_attempts_message: str,
    ) -> BoundRetryer:
        return BoundRetryer(self, retry_on_exceptions, no_more_attempts_message)

    async def __gen_attempts(self, errors: list[Exception]) -> t.AsyncIterable[Attempt]:
        retries_limit = self.__options.retries_limit or 0

        attempt_delay_factory = self.__provide_attempt_delay_factory()

        # i == 0 is a first attempt, it's not a retry yet
        for i in count():
            if i > retries_limit:
                return

            delay = attempt_delay_factory(i)
            if delay:
                await asyncio.sleep(delay)

            yield Attempt(i, errors)

    def __provide_attempt_delay_factory(self) -> t.Callable[[int], float | None]:
        if self.__options.retry_delay is None:

            def get_no_delay(_: int) -> None:
                return None

            return get_no_delay

        mode = self.__options.retry_delay_mode if self.__options.retry_delay_mode is not None else "multiplier"
        base_delay = self.__options.retry_delay.total_seconds()
        clamp_delay = self.__get_clamp_delay()

        if mode == "constant":

            def get_const_delay(_: int) -> float:
                return base_delay

            return get_const_delay

        elif mode == "multiplier":

            def get_mul_delay(i: int) -> float:
                return clamp_delay(base_delay * i)

            return get_mul_delay

        elif mode == "exponential":

            def get_exp_delay(i: int) -> float:
                return clamp_delay(base_delay**i)

            return get_exp_delay

        else:
            t.assert_never(mode)

    def __get_clamp_delay(self) -> t.Callable[[float], float]:
        if self.__options.retry_max_delay is not None:
            max_delay = self.__options.retry_max_delay.total_seconds()

            def clamp_delay(value: float) -> float:
                return min(value, max_delay)

        else:

            def clamp_delay(value: float) -> float:
                return value

        return clamp_delay



from __future__ import annotations

from brokrpc.message import Message, PackedMessage, UnpackedMessage
from brokrpc.model import BrokRPCError

type Request[T] = Message[T]
type BinaryRequest = Message[bytes]
type PackedRequest[T] = PackedMessage[T]
type UnpackedRequest[T] = UnpackedMessage[T]

type Response[T] = Message[T]
type BinaryResponse = Message[bytes]
type PackedResponse[T] = PackedMessage[T]
type UnpackedResponse[T] = UnpackedMessage[T]


class ServerError(BrokRPCError):
    pass


class HandlerError(ServerError):
    pass


class ClientError(BrokRPCError):
    pass


class CallerError(ClientError):
    pass



import builtins
import google.protobuf.any_pb2
import google.protobuf.message
import google.rpc.code_pb2
import typing

class Error(google.protobuf.message.Message):

    def __init__(self, *, message: typing.Optional[builtins.str]=None, details: typing.Sequence[google.protobuf.any_pb2.Any]) -> None:...

    @builtins.property
    def message(self) -> builtins.str:...

    @builtins.property
    def details(self) -> typing.Sequence[google.protobuf.any_pb2.Any]:...

    def HasField(self, field_name: typing.Literal['message']) -> builtins.bool:...

    def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

class UnaryRequest(google.protobuf.message.Message):

    def __init__(self, *, uuid: builtins.bytes, payload: google.protobuf.any_pb2.Any) -> None:...

    @builtins.property
    def uuid(self) -> builtins.bytes:...

    @builtins.property
    def payload(self) -> google.protobuf.any_pb2.Any:...

    def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

    def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

class UnaryResponse(google.protobuf.message.Message):

    def __init__(self, *, uuid: builtins.bytes, code: google.rpc.code_pb2.Code, payload: typing.Optional[google.protobuf.any_pb2.Any]=None, error: typing.Optional[Error]=None) -> None:...

    @builtins.property
    def uuid(self) -> builtins.bytes:...

    @builtins.property
    def code(self) -> google.rpc.code_pb2.Code:...

    @builtins.property
    def payload(self) -> google.protobuf.any_pb2.Any:...

    @builtins.property
    def error(self) -> Error:...

    def HasField(self, field_name: typing.Literal['payload', 'error']) -> builtins.bool:...

    def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

class StreamRequest(google.protobuf.message.Message):

    class Start(google.protobuf.message.Message):

        def __init__(self) -> None:...

        def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

        def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

    class Part(google.protobuf.message.Message):

        def __init__(self, *, payloads: typing.Sequence[google.protobuf.any_pb2.Any]) -> None:...

        @builtins.property
        def payloads(self) -> typing.Sequence[google.protobuf.any_pb2.Any]:...

        def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

        def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

    class End(google.protobuf.message.Message):

        def __init__(self) -> None:...

        def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

        def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

    def __init__(self, *, uuid: builtins.bytes, index: builtins.int, start: typing.Optional[StreamRequest.Start]=None, part: typing.Optional[StreamRequest.Part]=None, end: typing.Optional[StreamRequest.End]=None) -> None:...

    @builtins.property
    def uuid(self) -> builtins.bytes:...

    @builtins.property
    def index(self) -> builtins.int:...

    @builtins.property
    def start(self) -> StreamRequest.Start:...

    @builtins.property
    def part(self) -> StreamRequest.Part:...

    @builtins.property
    def end(self) -> StreamRequest.End:...

    def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

    def WhichOneof(self, oneof_group: typing.Literal['event']) -> typing.Optional[typing.Literal['start', 'part', 'end']]:...

class StreamResponse(google.protobuf.message.Message):

    class Start(google.protobuf.message.Message):

        def __init__(self) -> None:...

        def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

        def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

    class Part(google.protobuf.message.Message):

        def __init__(self, *, payloads: typing.Sequence[google.protobuf.any_pb2.Any]) -> None:...

        @builtins.property
        def payloads(self) -> typing.Sequence[google.protobuf.any_pb2.Any]:...

        def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

        def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

    class End(google.protobuf.message.Message):

        def __init__(self, *, code: google.rpc.code_pb2.Code, error: typing.Optional[Error]=None) -> None:...

        @builtins.property
        def code(self) -> google.rpc.code_pb2.Code:...

        @builtins.property
        def error(self) -> Error:...

        def HasField(self, field_name: typing.Literal['error']) -> builtins.bool:...

        def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

    def __init__(self, *, uuid: builtins.bytes, index: builtins.int, start: typing.Optional[StreamResponse.Start]=None, part: typing.Optional[StreamResponse.Part]=None, end: typing.Optional[StreamResponse.End]=None) -> None:...

    @builtins.property
    def uuid(self) -> builtins.bytes:...

    @builtins.property
    def index(self) -> builtins.int:...

    @builtins.property
    def start(self) -> StreamResponse.Start:...

    @builtins.property
    def part(self) -> StreamResponse.Part:...

    @builtins.property
    def end(self) -> StreamResponse.End:...

    def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

    def WhichOneof(self, oneof_group: typing.Literal['event']) -> typing.Optional[typing.Literal['start', 'part', 'end']]:...


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: brokrpc/spec/v1/amqp.proto
# Protobuf Python Version: 5.27.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    1,
    '',
    'brokrpc/spec/v1/amqp.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1a\x62rokrpc/spec/v1/amqp.proto\x12\x0f\x62rokrpc.spec.v1\x1a google/protobuf/descriptor.proto\"I\n\x11\x41rgumentListValue\x12\x34\n\x05items\x18\x01 \x03(\x0b\x32\x1e.brokrpc.spec.v1.ArgumentValueR\x05items\"\xb0\x01\n\x10\x41rgumentMapValue\x12\x42\n\x05items\x18\x01 \x03(\x0b\x32,.brokrpc.spec.v1.ArgumentMapValue.ItemsEntryR\x05items\x1aX\n\nItemsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32\x1e.brokrpc.spec.v1.ArgumentValueR\x05value:\x02\x38\x01\"\xc2\x02\n\rArgumentValue\x12\x1f\n\nnull_value\x18\x01 \x01(\x08H\x00R\tnullValue\x12\x1f\n\nbool_value\x18\x02 \x01(\x08H\x00R\tboolValue\x12\x1d\n\tint_value\x18\x03 \x01(\x03H\x00R\x08intValue\x12!\n\x0b\x66loat_value\x18\x04 \x01(\x02H\x00R\nfloatValue\x12\x1d\n\tstr_value\x18\x05 \x01(\tH\x00R\x08strValue\x12\x43\n\nlist_value\x18\x06 \x01(\x0b\x32\".brokrpc.spec.v1.ArgumentListValueH\x00R\tlistValue\x12@\n\tmap_value\x18\x07 \x01(\x0b\x32!.brokrpc.spec.v1.ArgumentMapValueH\x00R\x08mapValueB\x07\n\x05value\"\x82\x03\n\x0f\x45xchangeOptions\x12\x17\n\x04name\x18\x01 \x01(\tH\x00R\x04name\x88\x01\x01\x12\x36\n\x04type\x18\x02 \x01(\x0e\x32\x1d.brokrpc.spec.v1.ExchangeTypeH\x01R\x04type\x88\x01\x01\x12\x1d\n\x07\x64urable\x18\x03 \x01(\x08H\x02R\x07\x64urable\x88\x01\x01\x12$\n\x0b\x61uto_delete\x18\x04 \x01(\x08H\x03R\nautoDelete\x88\x01\x01\x12M\n\targuments\x18\x05 \x03(\x0b\x32/.brokrpc.spec.v1.ExchangeOptions.ArgumentsEntryR\targuments\x1a\\\n\x0e\x41rgumentsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32\x1e.brokrpc.spec.v1.ArgumentValueR\x05value:\x02\x38\x01\x42\x07\n\x05_nameB\x07\n\x05_typeB\n\n\x08_durableB\x0e\n\x0c_auto_delete\"\xec\x02\n\x0cQueueOptions\x12\x17\n\x04name\x18\x01 \x01(\tH\x00R\x04name\x88\x01\x01\x12\x1d\n\x07\x64urable\x18\x02 \x01(\x08H\x01R\x07\x64urable\x88\x01\x01\x12!\n\texclusive\x18\x03 \x01(\x08H\x02R\texclusive\x88\x01\x01\x12$\n\x0b\x61uto_delete\x18\x04 \x01(\x08H\x03R\nautoDelete\x88\x01\x01\x12J\n\targuments\x18\x05 \x03(\x0b\x32,.brokrpc.spec.v1.QueueOptions.ArgumentsEntryR\targuments\x1a\\\n\x0e\x41rgumentsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32\x1e.brokrpc.spec.v1.ArgumentValueR\x05value:\x02\x38\x01\x42\x07\n\x05_nameB\n\n\x08_durableB\x0c\n\n_exclusiveB\x0e\n\x0c_auto_delete*\x94\x01\n\x0c\x45xchangeType\x12\x1d\n\x19\x45XCHANGE_TYPE_UNSPECIFIED\x10\x00\x12\x18\n\x14\x45XCHANGE_TYPE_DIRECT\x10\x01\x12\x18\n\x14\x45XCHANGE_TYPE_FANOUT\x10\x02\x12\x17\n\x13\x45XCHANGE_TYPE_TOPIC\x10\x03\x12\x18\n\x14\x45XCHANGE_TYPE_HEADER\x10\x04:a\n\x08\x65xchange\x12\x1f.google.protobuf.ServiceOptions\x18\xa8, \x01(\x0b\x32 .brokrpc.spec.v1.ExchangeOptionsR\x08\x65xchange\x88\x01\x01:W\n\x05queue\x12\x1e.google.protobuf.MethodOptions\x18\xa8, \x01(\x0b\x32\x1d.brokrpc.spec.v1.QueueOptionsR\x05queue\x88\x01\x01\x42~\n\x13\x63om.brokrpc.spec.v1B\tAmqpProtoP\x01\xa2\x02\x03\x42SX\xaa\x02\x0f\x42rokrpc.Spec.V1\xca\x02\x0f\x42rokrpc\\Spec\\V1\xe2\x02\x1b\x42rokrpc\\Spec\\V1\\GPBMetadata\xea\x02\x11\x42rokrpc::Spec::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'brokrpc.spec.v1.amqp_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\023com.brokrpc.spec.v1B\tAmqpProtoP\001\242\002\003BSX\252\002\017Brokrpc.Spec.V1\312\002\017Brokrpc\\Spec\\V1\342\002\033Brokrpc\\Spec\\V1\\GPBMetadata\352\002\021Brokrpc::Spec::V1'
  _globals['_ARGUMENTMAPVALUE_ITEMSENTRY']._loaded_options = None
  _globals['_ARGUMENTMAPVALUE_ITEMSENTRY']._serialized_options = b'8\001'
  _globals['_EXCHANGEOPTIONS_ARGUMENTSENTRY']._loaded_options = None
  _globals['_EXCHANGEOPTIONS_ARGUMENTSENTRY']._serialized_options = b'8\001'
  _globals['_QUEUEOPTIONS_ARGUMENTSENTRY']._loaded_options = None
  _globals['_QUEUEOPTIONS_ARGUMENTSENTRY']._serialized_options = b'8\001'
  _globals['_EXCHANGETYPE']._serialized_start=1417
  _globals['_EXCHANGETYPE']._serialized_end=1565
  _globals['_ARGUMENTLISTVALUE']._serialized_start=81
  _globals['_ARGUMENTLISTVALUE']._serialized_end=154
  _globals['_ARGUMENTMAPVALUE']._serialized_start=157
  _globals['_ARGUMENTMAPVALUE']._serialized_end=333
  _globals['_ARGUMENTMAPVALUE_ITEMSENTRY']._serialized_start=245
  _globals['_ARGUMENTMAPVALUE_ITEMSENTRY']._serialized_end=333
  _globals['_ARGUMENTVALUE']._serialized_start=336
  _globals['_ARGUMENTVALUE']._serialized_end=658
  _globals['_EXCHANGEOPTIONS']._serialized_start=661
  _globals['_EXCHANGEOPTIONS']._serialized_end=1047
  _globals['_EXCHANGEOPTIONS_ARGUMENTSENTRY']._serialized_start=909
  _globals['_EXCHANGEOPTIONS_ARGUMENTSENTRY']._serialized_end=1001
  _globals['_QUEUEOPTIONS']._serialized_start=1050
  _globals['_QUEUEOPTIONS']._serialized_end=1414
  _globals['_QUEUEOPTIONS_ARGUMENTSENTRY']._serialized_start=909
  _globals['_QUEUEOPTIONS_ARGUMENTSENTRY']._serialized_end=1001
# @@protoc_insertion_point(module_scope)



from brokrpc.abc import Serializer


class IdentSerializer[T](Serializer[T, T]):
    def dump_message(self, message: T) -> T:
        return message

    def load_message(self, message: T) -> T:
        return message



import builtins
import enum
import google.protobuf.message
import typing

class ExchangeType(enum.IntEnum):
    EXCHANGE_TYPE_UNSPECIFIED = 0
    EXCHANGE_TYPE_DIRECT = 1
    EXCHANGE_TYPE_FANOUT = 2
    EXCHANGE_TYPE_TOPIC = 3
    EXCHANGE_TYPE_HEADER = 4

class ArgumentListValue(google.protobuf.message.Message):

    def __init__(self, *, items: typing.Sequence[ArgumentValue]) -> None:...

    @builtins.property
    def items(self) -> typing.Sequence[ArgumentValue]:...

    def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

    def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

class ArgumentMapValue(google.protobuf.message.Message):

    def __init__(self, *, items: typing.Mapping[builtins.str, ArgumentValue]) -> None:...

    @builtins.property
    def items(self) -> typing.Mapping[builtins.str, ArgumentValue]:...

    def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

    def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

class ArgumentValue(google.protobuf.message.Message):

    def __init__(self, *, null_value: typing.Optional[builtins.bool]=None, bool_value: typing.Optional[builtins.bool]=None, int_value: typing.Optional[builtins.int]=None, float_value: typing.Optional[builtins.float]=None, str_value: typing.Optional[builtins.str]=None, list_value: typing.Optional[ArgumentListValue]=None, map_value: typing.Optional[ArgumentMapValue]=None) -> None:...

    @builtins.property
    def null_value(self) -> builtins.bool:...

    @builtins.property
    def bool_value(self) -> builtins.bool:...

    @builtins.property
    def int_value(self) -> builtins.int:...

    @builtins.property
    def float_value(self) -> builtins.float:...

    @builtins.property
    def str_value(self) -> builtins.str:...

    @builtins.property
    def list_value(self) -> ArgumentListValue:...

    @builtins.property
    def map_value(self) -> ArgumentMapValue:...

    def HasField(self, field_name: typing.NoReturn) -> typing.NoReturn:...

    def WhichOneof(self, oneof_group: typing.Literal['value']) -> typing.Optional[typing.Literal['null_value', 'bool_value', 'int_value', 'float_value', 'str_value', 'list_value', 'map_value']]:...

class ExchangeOptions(google.protobuf.message.Message):

    def __init__(self, *, name: typing.Optional[builtins.str]=None, type: typing.Optional[ExchangeType]=None, durable: typing.Optional[builtins.bool]=None, auto_delete: typing.Optional[builtins.bool]=None, arguments: typing.Mapping[builtins.str, ArgumentValue]) -> None:...

    @builtins.property
    def name(self) -> builtins.str:...

    @builtins.property
    def type(self) -> ExchangeType:...

    @builtins.property
    def durable(self) -> builtins.bool:...

    @builtins.property
    def auto_delete(self) -> builtins.bool:...

    @builtins.property
    def arguments(self) -> typing.Mapping[builtins.str, ArgumentValue]:...

    def HasField(self, field_name: typing.Literal['name', 'type', 'durable', 'auto_delete']) -> builtins.bool:...

    def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...

class QueueOptions(google.protobuf.message.Message):

    def __init__(self, *, name: typing.Optional[builtins.str]=None, durable: typing.Optional[builtins.bool]=None, exclusive: typing.Optional[builtins.bool]=None, auto_delete: typing.Optional[builtins.bool]=None, arguments: typing.Mapping[builtins.str, ArgumentValue]) -> None:...

    @builtins.property
    def name(self) -> builtins.str:...

    @builtins.property
    def durable(self) -> builtins.bool:...

    @builtins.property
    def exclusive(self) -> builtins.bool:...

    @builtins.property
    def auto_delete(self) -> builtins.bool:...

    @builtins.property
    def arguments(self) -> typing.Mapping[builtins.str, ArgumentValue]:...

    def HasField(self, field_name: typing.Literal['name', 'durable', 'exclusive', 'auto_delete']) -> builtins.bool:...

    def WhichOneof(self, oneof_group: typing.NoReturn) -> typing.NoReturn:...


import abc
import typing as t

from brokrpc.message import BinaryMessage
from brokrpc.model import ConsumerResult, PublisherResult
from brokrpc.options import BindingOptions, PublisherOptions


class Publisher[U, V](metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def publish(self, message: U) -> V:
        raise NotImplementedError


class PublisherMiddleware[S, U, V](metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def publish(self, inner: S, message: U) -> V:
        raise NotImplementedError


class Consumer[U, V](metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def consume(self, message: U) -> V:
        raise NotImplementedError


class ConsumerMiddleware[S, U, V](metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def consume(self, inner: S, message: U) -> V:
        raise NotImplementedError


class BoundConsumer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def is_alive(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def get_options(self) -> BindingOptions:
        raise NotImplementedError


class Serializer[A, B](metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def dump_message(self, message: A) -> B:
        raise NotImplementedError

    @abc.abstractmethod
    def load_message(self, message: B) -> A:
        raise NotImplementedError


type BinaryPublisher = Publisher[BinaryMessage, PublisherResult]
type BinaryConsumer = Consumer[BinaryMessage, ConsumerResult]


class BrokerDriver(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def provide_publisher(self, options: PublisherOptions | None = None) -> t.AsyncContextManager[BinaryPublisher]:
        raise NotImplementedError

    @abc.abstractmethod
    def bind_consumer(self, consumer: BinaryConsumer, options: BindingOptions) -> t.AsyncContextManager[BoundConsumer]:
        raise NotImplementedError



import typing as t

from google.protobuf.any_pb2 import Any
from google.protobuf.json_format import Error, MessageToDict, MessageToJson
from google.protobuf.message import DecodeError, EncodeError
from google.protobuf.message import Message as ProtobufMessage

from brokrpc.abc import Serializer
from brokrpc.message import BinaryMessage, Message, PackedMessage, UnpackedMessage
from brokrpc.model import SerializerDumpError, SerializerLoadError
from brokrpc.rpc.abc import RPCSerializer
from brokrpc.rpc.model import BinaryRequest, BinaryResponse, Request, Response
from brokrpc.serializer.json import JSONSerializer


class ProtobufSerializer[T: ProtobufMessage](Serializer[Message[T], Message[bytes]]):
    __CONTENT_TYPE: t.Final[str] = "application/protobuf"

    def __init__(self, message_type: type[T]) -> None:
        self.__message_type = message_type

    def dump_message(self, message: Message[T]) -> PackedMessage[bytes]:
        assert isinstance(message.body, self.__message_type)

        try:
            body = message.body.SerializeToString()

        except EncodeError as err:
            details = f"can't encode protobuf message: {err}"
            raise SerializerDumpError(details, message) from err

        return PackedMessage(
            original=message,
            body=body,
            content_type=self.__CONTENT_TYPE,
            content_encoding=None,
            message_type=message.body.DESCRIPTOR.full_name,
        )

    def load_message(self, message: BinaryMessage) -> UnpackedMessage[T]:
        if message.content_type != self.__CONTENT_TYPE:
            details = f"invalid content type: {message.content_type}"
            raise SerializerLoadError(details, message)

        if message.message_type != self.__message_type.DESCRIPTOR.full_name:
            details = f"invalid message type: {message.message_type}"
            raise SerializerLoadError(details, message)

        try:
            body = self.__message_type.FromString(message.body)

        except DecodeError as err:
            details = f"can't decode protobuf message: {err}"
            raise SerializerLoadError(details, message) from err

        return UnpackedMessage(
            original=message,
            body=body,
        )


class DictProtobufSerializer[T: ProtobufMessage](Serializer[Message[dict[str, object]], Message[bytes]]):
    def __init__(self, message_type: type[T]) -> None:
        self.__message_type = message_type
        self.__proto = ProtobufSerializer(message_type)

    def dump_message(self, message: Message[dict[str, object]]) -> Message[bytes]:
        try:
            payload = self.__message_type(**message.body)

        except (TypeError, ValueError) as err:
            details = "can't construct protobuf message"
            raise SerializerDumpError(details, message.body) from err

        return self.__proto.dump_message(PackedMessage(body=payload, original=message))

    def load_message(self, message: Message[bytes]) -> Message[dict[str, object]]:
        protobuf_message = self.__proto.load_message(message)

        try:
            dict_payload = MessageToDict(protobuf_message.body, preserving_proto_field_name=True)

        except Error as err:
            details = "can't covert protobuf message to dict"
            raise SerializerLoadError(details, protobuf_message.body) from err

        return UnpackedMessage(
            original=protobuf_message,
            body=dict_payload,
        )


class JSONProtobufSerializer[T: ProtobufMessage](Serializer[Message[bytes], Message[bytes]]):
    def __init__(self, message_type: type[T], encoding: str | None = None) -> None:
        self.__message_type = message_type
        self.__json = JSONSerializer(encoding=encoding)
        self.__proto = ProtobufSerializer(message_type)

    def dump_message(self, message: Message[bytes]) -> Message[bytes]:
        try:
            json_message = self.__json.load_message(
                PackedMessage(
                    body=message.body,
                    content_type="application/json",
                    content_encoding=None,
                    message_type=None,
                    original=message,
                )
            )

        except SerializerLoadError as err:
            details = "can't load json message"
            raise SerializerDumpError(details, message) from err

        if not self.__check_json_message(json_message.body):
            details = "can't dump non object json to protobuf"
            raise SerializerDumpError(details, json_message)

        try:
            payload = self.__message_type(**json_message.body)

        except (TypeError, ValueError) as err:
            details = "can't construct protobuf message"
            raise SerializerDumpError(details, message.body) from err

        return self.__proto.dump_message(PackedMessage(body=payload, original=message))

    def load_message(self, message: Message[bytes]) -> Message[bytes]:
        protobuf_message = self.__proto.load_message(message)

        try:
            json_payload = MessageToJson(protobuf_message.body, preserving_proto_field_name=True, indent=None)

        except Error as err:
            details = "can't covert protobuf message to json"
            raise SerializerLoadError(details, protobuf_message.body) from err

        return UnpackedMessage(
            original=protobuf_message,
            body=json_payload.encode(),
        )

    def __check_json_message(self, obj: object) -> t.TypeGuard[dict[str, object]]:
        return isinstance(obj, dict) and all(isinstance(key, str) for key in obj)


class RPCProtobufSerializer[U: ProtobufMessage, V: ProtobufMessage](RPCSerializer[U, V]):
    def __init__(self, request_type: type[U], response_type: type[V]) -> None:
        self.__request = ProtobufSerializer(request_type)
        self.__response = ProtobufSerializer(response_type)

    def dump_unary_request(self, request: Request[U]) -> BinaryRequest:
        return self.__request.dump_message(request)

    def load_unary_request(self, request: BinaryRequest) -> Request[U]:
        return self.__request.load_message(request)

    def dump_unary_response(self, response: Response[V]) -> BinaryResponse:
        return self.__response.dump_message(response)

    def load_unary_response(self, response: BinaryResponse) -> Response[V]:
        return self.__response.load_message(response)


def pack_any(msg: ProtobufMessage) -> Any:
    result = Any()
    result.Pack(msg)
    return result


def unpack_any[T: ProtobufMessage](msg_type: type[T], value: Any) -> tuple[bool, T]:
    msg = msg_type()
    return value.Unpack(msg), msg



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: brokrpc/spec/v1/call.proto
# Protobuf Python Version: 5.27.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    1,
    '',
    'brokrpc/spec/v1/call.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.rpc import code_pb2 as google_dot_rpc_dot_code__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1a\x62rokrpc/spec/v1/call.proto\x12\x0f\x62rokrpc.spec.v1\x1a\x19google/protobuf/any.proto\x1a\x15google/rpc/code.proto\"b\n\x05\x45rror\x12\x1d\n\x07message\x18\x01 \x01(\tH\x00R\x07message\x88\x01\x01\x12.\n\x07\x64\x65tails\x18\x02 \x03(\x0b\x32\x14.google.protobuf.AnyR\x07\x64\x65tailsB\n\n\x08_message\"R\n\x0cUnaryRequest\x12\x12\n\x04uuid\x18\x01 \x01(\x0cR\x04uuid\x12.\n\x07payload\x18\x02 \x01(\x0b\x32\x14.google.protobuf.AnyR\x07payload\"\xc7\x01\n\rUnaryResponse\x12\x12\n\x04uuid\x18\x01 \x01(\x0cR\x04uuid\x12$\n\x04\x63ode\x18\x02 \x01(\x0e\x32\x10.google.rpc.CodeR\x04\x63ode\x12\x33\n\x07payload\x18\x03 \x01(\x0b\x32\x14.google.protobuf.AnyH\x00R\x07payload\x88\x01\x01\x12\x31\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x16.brokrpc.spec.v1.ErrorH\x01R\x05\x65rror\x88\x01\x01\x42\n\n\x08_payloadB\x08\n\x06_error\"\xbd\x02\n\rStreamRequest\x12\x12\n\x04uuid\x18\x01 \x01(\x0cR\x04uuid\x12\x14\n\x05index\x18\x02 \x01(\x04R\x05index\x12<\n\x05start\x18\x03 \x01(\x0b\x32$.brokrpc.spec.v1.StreamRequest.StartH\x00R\x05start\x12\x39\n\x04part\x18\x04 \x01(\x0b\x32#.brokrpc.spec.v1.StreamRequest.PartH\x00R\x04part\x12\x36\n\x03\x65nd\x18\x05 \x01(\x0b\x32\".brokrpc.spec.v1.StreamRequest.EndH\x00R\x03\x65nd\x1a\x07\n\x05Start\x1a\x38\n\x04Part\x12\x30\n\x08payloads\x18\x01 \x03(\x0b\x32\x14.google.protobuf.AnyR\x08payloads\x1a\x05\n\x03\x45ndB\x07\n\x05\x65vent\"\xa4\x03\n\x0eStreamResponse\x12\x12\n\x04uuid\x18\x01 \x01(\x0cR\x04uuid\x12\x14\n\x05index\x18\x02 \x01(\x04R\x05index\x12=\n\x05start\x18\x03 \x01(\x0b\x32%.brokrpc.spec.v1.StreamResponse.StartH\x00R\x05start\x12:\n\x04part\x18\x04 \x01(\x0b\x32$.brokrpc.spec.v1.StreamResponse.PartH\x00R\x04part\x12\x37\n\x03\x65nd\x18\x05 \x01(\x0b\x32#.brokrpc.spec.v1.StreamResponse.EndH\x00R\x03\x65nd\x1a\x07\n\x05Start\x1a\x38\n\x04Part\x12\x30\n\x08payloads\x18\x01 \x03(\x0b\x32\x14.google.protobuf.AnyR\x08payloads\x1ah\n\x03\x45nd\x12$\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x10.google.rpc.CodeR\x04\x63ode\x12\x31\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x16.brokrpc.spec.v1.ErrorH\x00R\x05\x65rror\x88\x01\x01\x42\x08\n\x06_errorB\x07\n\x05\x65ventB~\n\x13\x63om.brokrpc.spec.v1B\tCallProtoP\x01\xa2\x02\x03\x42SX\xaa\x02\x0f\x42rokrpc.Spec.V1\xca\x02\x0f\x42rokrpc\\Spec\\V1\xe2\x02\x1b\x42rokrpc\\Spec\\V1\\GPBMetadata\xea\x02\x11\x42rokrpc::Spec::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'brokrpc.spec.v1.call_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\023com.brokrpc.spec.v1B\tCallProtoP\001\242\002\003BSX\252\002\017Brokrpc.Spec.V1\312\002\017Brokrpc\\Spec\\V1\342\002\033Brokrpc\\Spec\\V1\\GPBMetadata\352\002\021Brokrpc::Spec::V1'
  _globals['_ERROR']._serialized_start=97
  _globals['_ERROR']._serialized_end=195
  _globals['_UNARYREQUEST']._serialized_start=197
  _globals['_UNARYREQUEST']._serialized_end=279
  _globals['_UNARYRESPONSE']._serialized_start=282
  _globals['_UNARYRESPONSE']._serialized_end=481
  _globals['_STREAMREQUEST']._serialized_start=484
  _globals['_STREAMREQUEST']._serialized_end=801
  _globals['_STREAMREQUEST_START']._serialized_start=720
  _globals['_STREAMREQUEST_START']._serialized_end=727
  _globals['_STREAMREQUEST_PART']._serialized_start=729
  _globals['_STREAMREQUEST_PART']._serialized_end=785
  _globals['_STREAMREQUEST_END']._serialized_start=787
  _globals['_STREAMREQUEST_END']._serialized_end=792
  _globals['_STREAMRESPONSE']._serialized_start=804
  _globals['_STREAMRESPONSE']._serialized_end=1224
  _globals['_STREAMRESPONSE_START']._serialized_start=720
  _globals['_STREAMRESPONSE_START']._serialized_end=727
  _globals['_STREAMRESPONSE_PART']._serialized_start=729
  _globals['_STREAMRESPONSE_PART']._serialized_end=785
  _globals['_STREAMRESPONSE_END']._serialized_start=1111
  _globals['_STREAMRESPONSE_END']._serialized_end=1215
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import logging
import typing as t
from datetime import timedelta

if t.TYPE_CHECKING:
    from brokrpc.options import ConsumerOptions

from brokrpc.abc import Consumer, ConsumerMiddleware, Publisher, PublisherMiddleware
from brokrpc.model import (
    ConsumerReject,
    ConsumerResult,
    ConsumerRetry,
    PublisherResult,
    SerializerDumpError,
    SerializerLoadError,
)
from brokrpc.stringify import to_str_obj


class PublisherMiddlewareWrapper[T, U, V](Publisher[U, V]):
    def __init__(self, inner: T, middleware: PublisherMiddleware[T, U, V]) -> None:
        self.__inner = inner
        self.__middleware = middleware

    def __str__(self) -> str:
        return to_str_obj(self, inner=self.__inner, middleware=self.__middleware)

    async def publish(self, message: U) -> V:
        return await self.__middleware.publish(self.__inner, message)


class ConsumerMiddlewareWrapper[T, U, V](Consumer[U, V]):
    def __init__(self, inner: T, middleware: ConsumerMiddleware[T, U, V]) -> None:
        self.__inner = inner
        self.__middleware = middleware

    def __str__(self) -> str:
        return to_str_obj(self, inner=self.__inner, middleware=self.__middleware)

    async def consume(self, message: U) -> V:
        return await self.__middleware.consume(self.__inner, message)


class AbortBadMessageMiddleware[T](
    PublisherMiddleware[Publisher[T, PublisherResult], T, PublisherResult],
    ConsumerMiddleware[Consumer[T, ConsumerResult], T, ConsumerResult],
):
    async def publish(self, inner: Publisher[T, PublisherResult], message: T) -> PublisherResult:
        try:
            result = await inner.publish(message)

        except SerializerDumpError as err:
            logging.warning("%s publisher serializer failed on %s, aborting", inner, message, exc_info=err)
            result = False

        return result

    async def consume(self, inner: Consumer[T, ConsumerResult], message: T) -> ConsumerResult:
        try:
            result = await inner.consume(message)

        except SerializerLoadError as err:
            logging.warning("%s consumer serializer failed on %s, aborting", inner, message, exc_info=err)
            result = ConsumerReject(
                reason=f"{inner} exception occurred: {err}",
            )

        return result


class RetryOnErrorConsumerMiddleware[T](ConsumerMiddleware[Consumer[T, ConsumerResult], T, ConsumerResult]):
    @classmethod
    def from_consume_options(cls, options: ConsumerOptions) -> RetryOnErrorConsumerMiddleware[T]:
        errors: tuple[type[Exception], ...]

        match options.retry_on_error:
            case True:
                errors = (Exception,)
            case False | None:
                errors = ()
            case tuple():
                errors = options.retry_on_error
            case error_type:
                errors = (error_type,)

        delay: timedelta | None

        match options.retry_delay:
            case float():
                delay = timedelta(seconds=options.retry_delay)
            case value:
                delay = value

        return cls(errors, delay)

    def __init__(
        self,
        retry_on_exceptions: tuple[type[Exception], ...],
        delay: timedelta | None = None,
    ) -> None:
        self.__retry_on_exceptions = retry_on_exceptions
        self.__delay = delay

    async def consume(self, inner: Consumer[T, ConsumerResult], message: T) -> ConsumerResult:
        try:
            result = await inner.consume(message)

        except self.__retry_on_exceptions as err:
            logging.warning("%s consumer failed on %s, retrying", inner, message, exc_info=err)
            result = ConsumerRetry(
                reason=f"{inner} exception occurred: {err}",
                delay=self.__delay,
            )

        return result



import typing as t
from json import JSONDecodeError, JSONDecoder, JSONEncoder

from brokrpc.abc import Serializer
from brokrpc.message import BinaryMessage, Message, PackedMessage, UnpackedMessage
from brokrpc.model import SerializerDumpError, SerializerLoadError
from brokrpc.rpc.abc import RPCSerializer
from brokrpc.rpc.model import BinaryRequest, Request, Response


class JSONSerializer(Serializer[Message[object], Message[bytes]], RPCSerializer[object, object]):
    __CONTENT_TYPE: t.Final[str] = "application/json"

    def __init__(
        self,
        encoder: JSONEncoder | None = None,
        decoder: JSONDecoder | None = None,
        encoding: str | None = None,
    ) -> None:
        self.__encoder = encoder if encoder is not None else JSONEncoder(indent=None, separators=(",", ":"))
        self.__decoder = decoder if decoder is not None else JSONDecoder()
        self.__encoding = encoding if encoding is not None else "utf-8"

    def dump_message(self, message: Message[object]) -> PackedMessage[bytes]:
        encoding = message.content_encoding if message.content_encoding is not None else self.__encoding
        try:
            body = self.__encoder.encode(message.body).encode(encoding)

        except ValueError as err:
            details = "can't encode json message"
            raise SerializerDumpError(details, message) from err

        return PackedMessage(
            original=message,
            body=body,
            content_type=self.__CONTENT_TYPE,
            content_encoding=encoding,
        )

    def load_message(self, message: BinaryMessage) -> UnpackedMessage[object]:
        obj = self.__load(message)
        return UnpackedMessage(message, obj)

    def dump_unary_request(self, request: Request[object]) -> BinaryRequest:
        return self.dump_message(request)

    def load_unary_request(self, request: BinaryRequest) -> Request[object]:
        return self.load_message(request)

    def dump_unary_response(self, response: Response[object]) -> BinaryMessage:
        return self.dump_message(response)

    def load_unary_response(self, response: BinaryMessage) -> Response[object]:
        return self.load_message(response)

    def __load(self, message: BinaryMessage) -> object:
        if message.content_type != self.__CONTENT_TYPE:
            details = f"invalid content type: {message.content_type}"
            raise SerializerLoadError(details, message)

        try:
            json_str = message.body.decode(
                message.content_encoding if message.content_encoding is not None else self.__encoding
            )

        except UnicodeDecodeError as err:
            details = f"can't decode binary string: {err}"
            raise SerializerLoadError(details, message) from err

        try:
            obj = self.__decoder.decode(json_str)

        except JSONDecodeError as err:
            details = f"can't decode json message: {err}"
            raise SerializerLoadError(details, message) from err

        return obj



from __future__ import annotations

import asyncio
import typing as t
import uuid
from contextlib import contextmanager


class WaiterStorage[T]:
    @classmethod
    @contextmanager
    def create(cls) -> t.Iterator[WaiterStorage[T]]:
        waiters: dict[uuid.UUID, asyncio.Future[T]] = {}
        try:
            yield cls(waiters)

        finally:
            for fut in waiters.values():
                if not fut.done():
                    fut.cancel()

            waiters.clear()

    def __init__(self, waiters: dict[uuid.UUID, asyncio.Future[T]]) -> None:
        self.__waiters = waiters

    @contextmanager
    def context(self) -> t.Iterator[tuple[str, asyncio.Future[T]]]:
        key = uuid.uuid4()
        fut = self.__waiters[key] = asyncio.Future[T]()

        try:
            yield key.hex, fut

        finally:
            if not fut.done():
                fut.cancel()

            self.__waiters.pop(key, None)

    def set_waiter(self, key: str, value: T | Exception) -> None:
        if isinstance(value, Exception):
            self.set_exception(key, value)

        else:
            self.set_result(key, value)

    def set_result(self, key: str, value: T) -> None:
        self.__waiters[uuid.UUID(hex=key)].set_result(value)

    def set_exception(self, key: str, value: Exception) -> None:
        self.__waiters[uuid.UUID(hex=key)].set_exception(value)

    def cancel(self, key: str, reason: str | None) -> None:
        self.__waiters[uuid.UUID(hex=key)].cancel(reason)



from __future__ import annotations

import asyncio
import typing as t

if t.TYPE_CHECKING:
    from concurrent.futures import Executor

    from brokrpc.message import BinaryMessage

from contextlib import AsyncExitStack
from functools import partial

from yarl import URL

from brokrpc.abc import (
    BinaryConsumer,
    BinaryPublisher,
    BoundConsumer,
    BrokerDriver,
    Consumer,
    ConsumerMiddleware,
    Publisher,
    PublisherMiddleware,
    Serializer,
)
from brokrpc.builder import ConsumerBuilder, PublisherBuilder, ident
from brokrpc.model import BrokerError, ConsumerResult, PublisherResult
from brokrpc.options import BindingOptions, BrokerOptions, ExchangeOptions, PublisherOptions, QueueOptions
from brokrpc.stringify import to_str_obj

type BrokerConnectOptions = str | URL | t.AsyncContextManager[BrokerDriver] | BrokerOptions


class BrokerIsNotConnectedError(BrokerError):
    pass


class Broker(t.AsyncContextManager["Broker"]):
    def __init__(
        self,
        options: BrokerConnectOptions,
        *,
        default_exchange: ExchangeOptions | None = None,
        default_queue: QueueOptions | None = None,
        default_publisher_middlewares: t.Sequence[PublisherMiddleware[BinaryPublisher, BinaryMessage, PublisherResult]]
        | None = None,
        default_consumer_middlewares: t.Sequence[ConsumerMiddleware[BinaryConsumer, BinaryMessage, ConsumerResult]]
        | None = None,
    ) -> None:
        self.__options = options
        self.__connect: t.Callable[[], t.AsyncContextManager[BrokerDriver]] = partial(connect, options)
        self.__default_exchange = default_exchange
        self.__default_queue = default_queue
        self.__default_publisher_middlewares: t.Sequence[
            PublisherMiddleware[BinaryPublisher, BinaryMessage, PublisherResult]
        ] = default_publisher_middlewares or ()
        self.__default_consumer_middlewares: t.Sequence[
            ConsumerMiddleware[BinaryConsumer, BinaryMessage, ConsumerResult]
        ] = default_consumer_middlewares or ()

        self.__stack = AsyncExitStack()
        self.__lock = asyncio.Lock()
        self.__opened = asyncio.Event()
        self.__driver: BrokerDriver | None = None

    def __str__(self) -> str:
        return to_str_obj(self, is_connected=self.is_connected, driver=self.__driver)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"options={self.__options!r}, "
            f"default_exchange={self.__default_exchange!r}, "
            f"default_queue={self.__default_queue!r}, "
            f"default_publisher_middlewares={self.__default_publisher_middlewares!r}, "
            f"default_consumer_middlewares={self.__default_consumer_middlewares!r}"
            ")"
        )

    async def __aenter__(self) -> t.Self:
        await self.connect()
        return self

    async def __aexit__(self, exc_type: object, exc_val: object, exc_tb: object) -> bool | None:
        await self.disconnect()
        return None

    async def connect(self) -> None:
        if self.__driver is not None:
            return

        async with self.__lock:
            if self.__driver is not None:
                return

            assert not self.is_connected
            self.__driver = await self.__stack.enter_async_context(self.__connect())
            self.__opened.set()
            assert self.is_connected

    async def disconnect(self) -> None:
        if self.__driver is None:
            return

        async with self.__lock:
            if self.__driver is None:
                return

            assert self.is_connected
            try:
                await self.__stack.aclose()

            finally:
                self.__driver = None
                self.__opened.clear()
                assert not self.is_connected

    @property
    def is_connected(self) -> bool:
        return self.__driver is not None and self.__opened.is_set()

    @t.overload
    def publisher[T](
        self,
        options: PublisherOptions | None = None,
    ) -> t.AsyncContextManager[BinaryPublisher]: ...

    @t.overload
    def publisher[T](
        self,
        options: PublisherOptions | None = None,
        *,
        serializer: t.Callable[[T], BinaryMessage] | Serializer[T, BinaryMessage],
    ) -> t.AsyncContextManager[Publisher[T, PublisherResult]]: ...

    def publisher[T](
        self,
        options: PublisherOptions | None = None,
        *,
        serializer: t.Callable[[T], BinaryMessage] | Serializer[T, BinaryMessage] | None = None,
    ) -> t.AsyncContextManager[BinaryPublisher] | t.AsyncContextManager[Publisher[T, PublisherResult]]:
        return self.builder_publisher().add_serializer(serializer).build(options)

    @t.overload
    def consumer[T](
        self,
        consumer: BinaryConsumer,
        options: BindingOptions,
    ) -> t.AsyncContextManager[BoundConsumer]: ...

    @t.overload
    def consumer[T](
        self,
        consumer: t.Callable[[BinaryMessage], t.Awaitable[ConsumerResult]],
        options: BindingOptions,
    ) -> t.AsyncContextManager[BoundConsumer]: ...

    @t.overload
    def consumer[T](
        self,
        consumer: t.Callable[[BinaryMessage], ConsumerResult],
        options: BindingOptions,
        *,
        executor: Executor | None = None,
    ) -> t.AsyncContextManager[BoundConsumer]: ...

    @t.overload
    def consumer[T](
        self,
        consumer: Consumer[T, ConsumerResult],
        options: BindingOptions,
        *,
        serializer: t.Callable[[BinaryMessage], T] | Serializer[T, BinaryMessage],
    ) -> t.AsyncContextManager[BoundConsumer]: ...

    @t.overload
    def consumer[T](
        self,
        consumer: t.Callable[[T], t.Awaitable[ConsumerResult]],
        options: BindingOptions,
        *,
        serializer: t.Callable[[BinaryMessage], T] | Serializer[T, BinaryMessage],
    ) -> t.AsyncContextManager[BoundConsumer]: ...

    @t.overload
    def consumer[T](
        self,
        consumer: t.Callable[[T], ConsumerResult],
        options: BindingOptions,
        *,
        serializer: t.Callable[[BinaryMessage], T] | Serializer[T, BinaryMessage],
        executor: Executor | None = None,
    ) -> t.AsyncContextManager[BoundConsumer]: ...

    def consumer[T](
        self,
        consumer: BinaryConsumer
        | t.Callable[[BinaryMessage], t.Awaitable[ConsumerResult]]
        | t.Callable[[BinaryMessage], ConsumerResult]
        | Consumer[T, ConsumerResult]
        | t.Callable[[T], t.Awaitable[ConsumerResult]]
        | t.Callable[[T], ConsumerResult],
        options: BindingOptions,
        serializer: t.Callable[[BinaryMessage], T] | Serializer[T, BinaryMessage] | None = None,
        executor: Executor | None = None,
    ) -> t.AsyncContextManager[BoundConsumer]:
        return (
            self.build_consumer()
            .add_serializer(serializer)
            # TODO: remove cast to any
            .build(t.cast(t.Any, consumer), options, executor=executor)
        )

    def builder_publisher(self) -> PublisherBuilder[BinaryMessage, PublisherResult]:
        return (
            PublisherBuilder(self.__provide_publisher, ident)
            .set_exchange(self.__default_exchange)
            .add_middlewares(*self.__default_publisher_middlewares)
        )

    def build_consumer(self) -> ConsumerBuilder[BinaryMessage, ConsumerResult]:
        return (
            ConsumerBuilder(self.__bind_consumer, ident)
            .set_exchange(self.__default_exchange)
            .set_queue(self.__default_queue)
            .add_middlewares(*self.__default_consumer_middlewares)
        )

    def __get_driver(self) -> BrokerDriver:
        if self.__driver is None:
            assert not self.is_connected
            raise BrokerIsNotConnectedError(self)

        assert self.is_connected
        return self.__driver

    def __provide_publisher(self, options: PublisherOptions | None) -> t.AsyncContextManager[BinaryPublisher]:
        return self.__get_driver().provide_publisher(options)

    def __bind_consumer(
        self, consumer: BinaryConsumer, options: BindingOptions
    ) -> t.AsyncContextManager[BoundConsumer]:
        return self.__get_driver().bind_consumer(consumer, options)


def connect(options: BrokerConnectOptions) -> t.AsyncContextManager[BrokerDriver]:
    if isinstance(options, t.AsyncContextManager):
        return options

    clean_options = options if isinstance(options, BrokerOptions) else parse_options(options)

    if clean_options.driver == "aiormq":
        from brokrpc.driver.aiormq import AiormqBrokerDriver

        return AiormqBrokerDriver.connect(clean_options)

    else:
        details = "unsupported driver"
        raise ValueError(details, clean_options)


def parse_options(url: str | URL) -> BrokerOptions:
    clean_url = URL(url) if isinstance(url, str) else url

    parts = clean_url.scheme.split("+", maxsplit=1)
    match parts:
        case (scheme, driver):
            return BrokerOptions(driver=driver, url=clean_url.with_scheme(scheme))

        case _:
            return BrokerOptions(driver="aiormq", url=clean_url)



from __future__ import annotations

import asyncio
import typing as t
from contextlib import asynccontextmanager
from dataclasses import replace
from datetime import UTC, datetime, timedelta
from uuid import uuid4

from aiormq import Channel, Connection, ProtocolSyntaxError, spec
from pamqp.common import FieldTable

from brokrpc.retry import Retryer
from brokrpc.stringify import to_str_obj

if t.TYPE_CHECKING:
    from aiormq.abc import DeliveredMessage

from brokrpc.abc import BinaryConsumer, BinaryPublisher, BoundConsumer, BrokerDriver, Publisher
from brokrpc.message import BinaryMessage, Message
from brokrpc.model import (
    ConsumerAck,
    ConsumerReject,
    ConsumerResult,
    ConsumerRetry,
    PublisherResult,
)
from brokrpc.options import (
    BindingOptions,
    BrokerOptions,
    ExchangeOptions,
    PublisherOptions,
    QOSOptions,
    QueueOptions,
)


class AiormqMessage(Message[bytes]):
    __slots__ = ("__impl",)

    def __init__(self, impl: DeliveredMessage) -> None:
        self.__impl = impl
        assert isinstance(self.__impl.routing_key, str)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__impl!r})"

    @property
    def body(self) -> bytes:
        return self.__impl.body

    @property
    def routing_key(self) -> str:
        assert isinstance(self.__impl.routing_key, str)
        return self.__impl.routing_key

    @property
    def exchange(self) -> str | None:
        return self.__impl.exchange

    @property
    def content_type(self) -> str | None:
        return self.__impl.header.properties.content_type

    @property
    def content_encoding(self) -> str | None:
        return self.__impl.header.properties.content_encoding

    @property
    def headers(self) -> t.Mapping[str, str] | None:
        return t.cast(t.Mapping[str, str], self.__impl.header.properties.headers)

    @property
    def delivery_mode(self) -> int | None:
        return self.__impl.header.properties.delivery_mode

    @property
    def priority(self) -> int | None:
        return self.__impl.header.properties.priority

    @property
    def correlation_id(self) -> str | None:
        return self.__impl.header.properties.correlation_id

    @property
    def reply_to(self) -> str | None:
        return self.__impl.header.properties.reply_to

    @property
    def timeout(self) -> timedelta | None:
        return _load_message_timeout(self.__impl.header.properties.expiration)

    @property
    def message_id(self) -> str | None:
        return self.__impl.header.properties.message_id

    @property
    def timestamp(self) -> datetime | None:
        return self.__impl.header.properties.timestamp

    @property
    def message_type(self) -> str | None:
        return self.__impl.header.properties.message_type

    @property
    def user_id(self) -> str | None:
        return self.__impl.header.properties.user_id

    @property
    def app_id(self) -> str | None:
        return self.__impl.header.properties.app_id


class AiormqPublisher(Publisher[BinaryMessage, PublisherResult]):
    def __init__(
        self,
        channel: Channel,
        options: PublisherOptions | None,
        message_id_gen: t.Callable[[], str | None],
        now: t.Callable[[], datetime | None],
    ) -> None:
        self.__channel = channel
        self.__options = options
        self.__message_id_gen = message_id_gen
        self.__now = now

    async def publish(self, message: BinaryMessage) -> PublisherResult:
        await self.__channel.basic_publish(
            body=message.body,
            exchange=message.exchange
            if message.exchange is not None
            else self.__options.name
            if self.__options is not None and self.__options.name is not None
            else "",
            routing_key=message.routing_key,
            mandatory=self.__options.mandatory
            if self.__options is not None and self.__options.mandatory is not None
            else False,
            properties=spec.Basic.Properties(
                content_type=message.content_type,
                content_encoding=message.content_encoding,
                headers=t.cast(FieldTable | None, message.headers),
                delivery_mode=message.delivery_mode,
                priority=message.priority,
                correlation_id=message.correlation_id,
                reply_to=message.reply_to,
                expiration=_dump_message_timeout(message.timeout),
                message_id=message.message_id if message.message_id is not None else self.__message_id_gen(),
                timestamp=message.timestamp if message.timestamp is not None else self.__now(),
                message_type=message.message_type,
                user_id=message.user_id,
                app_id=message.app_id,
            ),
        )

        return None


class AiormqConsumerCallback:
    def __init__(self, channel: Channel, inner: BinaryConsumer) -> None:
        self.__channel = channel
        self.__inner = inner

    async def __call__(self, aiormq_message: DeliveredMessage) -> None:
        assert isinstance(aiormq_message.delivery_tag, int)
        assert isinstance(aiormq_message.routing_key, str)

        message = AiormqMessage(aiormq_message)
        result = await self.__inner.consume(message)
        # TODO: check how aiormq handles exception
        await self.__handle_result(aiormq_message, result)

    async def __handle_result(self, message: DeliveredMessage, result: ConsumerResult) -> None:
        assert isinstance(message.delivery_tag, int)

        match result:
            case ConsumerAck() | True | None:
                await self.__channel.basic_ack(message.delivery_tag)

            case ConsumerReject() | False:
                await self.__channel.basic_reject(message.delivery_tag, requeue=False)

            case ConsumerRetry(delay=delay):
                if delay is not None:
                    await asyncio.sleep(delay.total_seconds())

                # TODO: AMQP returns message to the head of the queue, think about option to return it to the end of the
                #  queue (real requeue).
                await self.__channel.basic_reject(message.delivery_tag)

            case _:
                t.assert_never(result)


class AiormqBoundConsumer(BoundConsumer):
    def __init__(
        self,
        channel: Channel,
        callback: AiormqConsumerCallback,
        options: BindingOptions,
        consumer_tag: str,
    ) -> None:
        self.__channel = channel
        self.__callback = callback
        self.__options = options
        self.__consumer_tag = consumer_tag

    def is_alive(self) -> bool:
        return self.__consumer_tag in self.__channel.consumers

    def get_options(self) -> BindingOptions:
        return self.__options


class AiormqBrokerDriver(BrokerDriver):
    @classmethod
    @asynccontextmanager
    async def connect(cls, options: BrokerOptions) -> t.AsyncIterator[AiormqBrokerDriver]:
        conn = Connection(options.url)

        try:
            await (
                Retryer(options)
                .bind(
                    retry_on_exceptions=(ConnectionError, ProtocolSyntaxError),
                    no_more_attempts_message="can't connect to rabbitmq",
                )
                .do(conn.connect)
            )
            yield cls(conn)

        finally:
            await conn.close()

    def __init__(self, connection: Connection) -> None:
        self.__connection = connection

    def __str__(self) -> str:
        return to_str_obj(self, connection=self.__connection)

    @asynccontextmanager
    async def provide_publisher(self, options: PublisherOptions | None = None) -> t.AsyncIterator[BinaryPublisher]:
        async with self.__provide_channel(options) as channel:
            yield AiormqPublisher(channel, options, self.__gen_message_id, self.__get_now)

    @asynccontextmanager
    async def bind_consumer(self, consumer: BinaryConsumer, options: BindingOptions) -> t.AsyncIterator[BoundConsumer]:
        channel: Channel

        async with self.__provide_channel(options.queue) as channel:
            callback = AiormqConsumerCallback(channel, consumer)

            bindings = await self.__bind(channel, options)
            assert isinstance(bindings.queue, QueueOptions)
            assert isinstance(bindings.queue.name, str)

            consume_ok = await channel.basic_consume(
                queue=bindings.queue.name,
                consumer_callback=callback,
            )
            assert isinstance(consume_ok.consumer_tag, str)

            try:
                yield AiormqBoundConsumer(channel, callback, bindings, consume_ok.consumer_tag)

            finally:
                await channel.basic_cancel(consume_ok.consumer_tag)

    @asynccontextmanager
    async def __provide_channel(self, options: QOSOptions | None) -> t.AsyncIterator[Channel]:
        channel = await self.__connection.channel()
        assert isinstance(channel, Channel)

        try:
            if options is not None:
                await channel.basic_qos(
                    prefetch_count=options.prefetch_count,
                )

            yield channel

        finally:
            await channel.close()

    async def __setup_exchange(
        self,
        channel: Channel,
        options: ExchangeOptions | None,
    ) -> tuple[str, ExchangeOptions]:
        name = options.name if options is not None and options.name is not None else ""
        type_ = options.type if options is not None and options.type is not None else "direct"
        durable = options.durable if options is not None and options.durable is not None else False
        auto_delete = options.auto_delete if options is not None and options.auto_delete is not None else False
        arguments = options.arguments if options is not None and options.arguments is not None else None

        await channel.exchange_declare(
            exchange=name,
            exchange_type=type_,
            durable=durable,
            auto_delete=auto_delete,
            # TODO: remove ignore
            arguments=arguments,  # type: ignore[arg-type]
        )

        return name, ExchangeOptions(
            name=name,
            type=type_,
            durable=durable,
            auto_delete=auto_delete,
            arguments=arguments,
        )

    async def __setup_queue(
        self,
        channel: Channel,
        options: QueueOptions | None,
    ) -> tuple[str, QueueOptions]:
        name = options.name if options is not None and options.name is not None else ""
        durable = options.durable if options is not None and options.durable is not None else False
        exclusive = options.exclusive if options is not None and options.exclusive is not None else False
        auto_delete = options.auto_delete if options is not None and options.auto_delete is not None else False
        arguments = options.arguments if options is not None and options.arguments is not None else None

        declare_ok = await channel.queue_declare(
            queue=name,
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            # TODO: remove ignore
            arguments=arguments,  # type: ignore[arg-type]
        )
        assert isinstance(declare_ok.queue, str)

        return declare_ok.queue, QueueOptions(
            name=declare_ok.queue,
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            arguments=arguments,
        )

    async def __bind(self, channel: Channel, options: BindingOptions) -> BindingOptions:
        exchange_name, exchange_options = await self.__setup_exchange(channel, options.exchange)
        queue_name, queue_options = await self.__setup_queue(channel, options.queue)

        await asyncio.gather(
            *(
                channel.queue_bind(
                    exchange=exchange_name,
                    routing_key=binding_key,
                    queue=queue_name,
                )
                for binding_key in options.binding_keys
            )
        )

        return replace(options, exchange=exchange_options, queue=queue_options)

    def __gen_message_id(self) -> str:
        return uuid4().hex

    def __get_now(self) -> datetime:
        return datetime.now(tz=UTC)


def _load_message_timeout(value: str | None) -> timedelta | None:
    return timedelta(milliseconds=float(value)) if value else None


def _dump_message_timeout(value: timedelta | None) -> str | None:
    return f"{value.total_seconds() * 1000:.0f}" if value is not None else None



import typing as t
import uuid
from contextlib import asynccontextmanager
from datetime import timedelta

from brokrpc.broker import Broker
from brokrpc.options import BindingOptions, ExchangeOptions, QueueOptions, merge_options
from brokrpc.rpc.abc import Caller, CallerSerializer
from brokrpc.rpc.caller import RequestCaller
from brokrpc.rpc.handler import HandlerResponseConsumer
from brokrpc.rpc.model import Response
from brokrpc.rpc.storage import WaiterStorage


class Client:
    def __init__(self, broker: Broker) -> None:
        self.__broker = broker

    @asynccontextmanager
    async def unary_unary_caller[U, V](
        self,
        *,
        routing_key: str,
        serializer: CallerSerializer[U, V],
        exchange: ExchangeOptions | None = None,
        queue: QueueOptions | None = None,
        # NOTE: `timeout` parameter passed to server via message, so server can cancel the task if it is expired.
        timeout: timedelta | None = None,  # noqa: ASYNC109
    ) -> t.AsyncIterator[Caller[U, V]]:
        caller_id = uuid.uuid4()
        response_key = f"response.{routing_key}.{caller_id.hex}"
        binding = self.__get_binding_options(exchange, response_key, queue)

        with WaiterStorage[Response[V]].create() as storage:
            async with (
                self.__broker.consumer(
                    HandlerResponseConsumer(serializer.load_unary_response, storage),
                    binding,
                ),
                self.__broker.publisher() as requester,
            ):
                yield RequestCaller(
                    requester=requester,
                    exchange=binding.exchange,
                    routing_key=routing_key,
                    serializer=serializer.dump_unary_request,
                    reply_to=response_key,
                    storage=storage,
                    timeout=timeout,
                )

    def __get_binding_options(
        self,
        exchange: ExchangeOptions | None,
        response_key: str,
        queue: QueueOptions | None,
    ) -> BindingOptions:
        return BindingOptions(
            exchange=exchange,
            binding_keys=(response_key,),
            queue=merge_options(
                queue,
                QueueOptions(
                    name=response_key,
                    auto_delete=True,
                    exclusive=True,
                ),
            ),
        )



from __future__ import annotations

import inspect
import typing as t

if t.TYPE_CHECKING:
    from concurrent.futures import Executor

from contextlib import asynccontextmanager
from dataclasses import replace

from brokrpc.abc import (
    BinaryConsumer,
    BinaryPublisher,
    BoundConsumer,
    Consumer,
    ConsumerMiddleware,
    Publisher,
    PublisherMiddleware,
    Serializer,
)
from brokrpc.consumer import AsyncFuncConsumer, DecodingConsumer, SyncFuncConsumer
from brokrpc.middleware import ConsumerMiddlewareWrapper, PublisherMiddlewareWrapper
from brokrpc.options import BindingOptions, ExchangeOptions, PublisherOptions, QueueOptions, merge_options
from brokrpc.publisher import EncodingPublisher


def ident[T](obj: T) -> T:
    return obj


class PublisherBuilder[U, V]:
    def __init__(
        self,
        provider: t.Callable[[PublisherOptions | None], t.AsyncContextManager[BinaryPublisher]],
        wrapper: t.Callable[[BinaryPublisher], Publisher[U, V]],
        exchange: ExchangeOptions | None = None,
    ) -> None:
        self.__provider = provider
        self.__wrapper = wrapper
        self.__exchange = exchange

    def set_exchange(self, exchange: ExchangeOptions | None) -> PublisherBuilder[U, V]:
        return PublisherBuilder(self.__provider, self.__wrapper, merge_options(self.__exchange, exchange))

    @t.overload
    def add_serializer(self, serializer: None) -> PublisherBuilder[U, V]: ...

    @t.overload
    def add_serializer[U2](self, serializer: t.Callable[[U2], U]) -> PublisherBuilder[U2, V]: ...

    @t.overload
    def add_serializer[U2](self, serializer: Serializer[U2, U]) -> PublisherBuilder[U2, V]: ...

    def add_serializer[U2](
        self,
        serializer: t.Callable[[U2], U] | Serializer[U2, U] | None,
    ) -> PublisherBuilder[U2, V]:
        if serializer is None:
            return t.cast(PublisherBuilder[U2, V], self)

        inner_wrapper = self.__wrapper
        dump_message = serializer.dump_message if isinstance(serializer, Serializer) else serializer

        def wrapper(inner: BinaryPublisher) -> Publisher[U2, V]:
            return EncodingPublisher(inner_wrapper(inner), dump_message)

        return PublisherBuilder(self.__provider, wrapper, self.__exchange)

    def add_middleware[U2, V2](
        self,
        middleware: PublisherMiddleware[Publisher[U, V], U2, V2],
    ) -> PublisherBuilder[U2, V2]:
        inner_wrapper = self.__wrapper

        def wrapper(inner: BinaryPublisher) -> Publisher[U2, V2]:
            return PublisherMiddlewareWrapper(inner_wrapper(inner), middleware)

        return PublisherBuilder(self.__provider, wrapper, self.__exchange)

    def add_middlewares(
        self,
        *middlewares: PublisherMiddleware[Publisher[U, V], U, V],
    ) -> PublisherBuilder[U, V]:
        result = self

        for middleware in middlewares:
            result = result.add_middleware(middleware)

        return result

    @asynccontextmanager
    async def build(self, options: PublisherOptions | None = None) -> t.AsyncIterator[Publisher[U, V]]:
        async with self.__provider(self.__clear_options(options)) as publisher:
            yield self.__wrapper(publisher)

    def __clear_options(self, options: PublisherOptions | None) -> PublisherOptions | None:
        if self.__exchange is None:
            return options

        return PublisherOptions(
            prefetch_count=options.prefetch_count if options is not None else self.__exchange.prefetch_count,
            name=options.name if options is not None else self.__exchange.name,
            type=options.type if options is not None else self.__exchange.type,
            durable=options.durable if options is not None else self.__exchange.durable,
            auto_delete=options.auto_delete if options is not None else self.__exchange.auto_delete,
            arguments=options.arguments if options is not None else self.__exchange.arguments,
            mandatory=options.mandatory if options is not None else None,
        )


class ConsumerBuilder[U, V]:
    def __init__(
        self,
        binder: t.Callable[[BinaryConsumer, BindingOptions], t.AsyncContextManager[BoundConsumer]],
        wrapper: t.Callable[[Consumer[U, V]], BinaryConsumer],
        exchange: ExchangeOptions | None = None,
        queue: QueueOptions | None = None,
    ) -> None:
        self.__binder = binder
        self.__wrapper = wrapper
        self.__exchange = exchange
        self.__queue = queue

    def set_exchange(self, exchange: ExchangeOptions | None) -> ConsumerBuilder[U, V]:
        return ConsumerBuilder(self.__binder, self.__wrapper, merge_options(self.__exchange, exchange), self.__queue)

    def set_queue(self, queue: QueueOptions | None) -> ConsumerBuilder[U, V]:
        return ConsumerBuilder(self.__binder, self.__wrapper, self.__exchange, merge_options(self.__queue, queue))

    @t.overload
    def add_serializer(self, serializer: None) -> ConsumerBuilder[U, V]: ...

    @t.overload
    def add_serializer[U2](self, serializer: t.Callable[[U], U2]) -> ConsumerBuilder[U2, V]: ...

    @t.overload
    def add_serializer[U2](self, serializer: Serializer[U2, U]) -> ConsumerBuilder[U2, V]: ...

    def add_serializer[U2](self, serializer: t.Callable[[U], U2] | Serializer[U2, U] | None) -> ConsumerBuilder[U2, V]:
        if serializer is None:
            return t.cast(ConsumerBuilder[U2, V], self)

        inner_wrapper = self.__wrapper
        load_message = serializer.load_message if isinstance(serializer, Serializer) else serializer

        def wrapper(inner: Consumer[U2, V]) -> BinaryConsumer:
            return inner_wrapper(DecodingConsumer(inner, load_message))

        return ConsumerBuilder(self.__binder, wrapper, self.__exchange, self.__queue)

    def add_middleware(
        self,
        middleware: ConsumerMiddleware[Consumer[U, V], U, V],
    ) -> ConsumerBuilder[U, V]:
        inner_wrapper = self.__wrapper

        def wrapper(inner: Consumer[U, V]) -> BinaryConsumer:
            return inner_wrapper(ConsumerMiddlewareWrapper(inner, middleware))

        return ConsumerBuilder(self.__binder, wrapper, self.__exchange, self.__queue)

    def add_middlewares(
        self,
        *middlewares: ConsumerMiddleware[Consumer[U, V], U, V],
    ) -> ConsumerBuilder[U, V]:
        result = self

        for middleware in middlewares:
            result = result.add_middleware(middleware)

        return result

    @t.overload
    def build(
        self,
        inner: Consumer[U, V],
        options: BindingOptions,
    ) -> t.AsyncContextManager[BoundConsumer]: ...

    @t.overload
    def build(
        self,
        inner: t.Callable[[U], t.Coroutine[object, object, V]],
        options: BindingOptions,
    ) -> t.AsyncContextManager[BoundConsumer]: ...

    @t.overload
    def build(
        self,
        inner: t.Callable[[U], V],
        options: BindingOptions,
        *,
        executor: Executor | None = None,
    ) -> t.AsyncContextManager[BoundConsumer]: ...

    def build(
        self,
        inner: Consumer[U, V] | t.Callable[[U], t.Coroutine[object, object, V]] | t.Callable[[U], V],
        options: BindingOptions,
        executor: Executor | None = None,
    ) -> t.AsyncContextManager[BoundConsumer]:
        consumer: Consumer[U, V]

        match inner:
            case Consumer():
                consumer = inner

            case async_func if inspect.iscoroutinefunction(async_func):
                consumer = AsyncFuncConsumer(async_func)

            case sync_func if callable(sync_func):
                consumer = SyncFuncConsumer(t.cast(t.Callable[[U], V], sync_func), executor)

            case _:
                # FIXME: make `t.assert_never(inner)` work
                raise TypeError(inner)

        return self.__binder(self.__wrapper(consumer), self.__clear_binding_options(options))

    def __clear_binding_options(self, binding: BindingOptions) -> BindingOptions:
        if self.__exchange is None and self.__queue is None:
            return binding

        return replace(
            binding,
            exchange=merge_options(self.__exchange, binding.exchange),
            queue=merge_options(self.__queue, binding.queue),
        )



import typing as t


def to_str_obj(
    obj: object,
    type_vars: t.Sequence[type[object]] | None = None,
    **kwargs: object,
) -> str:
    parts = [f"{to_str_type(type(obj), type_vars)} object at {hex(id(obj))}"]
    for key, value in kwargs.items():
        parts.append(f"{key}={value}")

    return f"""<{"; ".join(parts)}>"""


def to_str_type(
    type_: type[object],
    type_vars: t.Sequence[type[object]] | None = None,
) -> str:
    value = f"{type_.__module__}.{type_.__name__}"
    return value if not type_vars else f"""{value}[{", ".join(tv.__name__ for tv in type_vars)}]"""



from __future__ import annotations

import typing as t
from dataclasses import dataclass

if t.TYPE_CHECKING:
    from datetime import timedelta

type PublisherResult = bool | None


@dataclass(frozen=True, kw_only=True)
class ConsumerAck:
    pass


@dataclass(frozen=True, kw_only=True)
class ConsumerReject:
    reason: str | None = None


@dataclass(frozen=True, kw_only=True)
class ConsumerRetry:
    reason: str | None = None
    delay: timedelta | None


type ConsumerResult = ConsumerAck | ConsumerReject | ConsumerRetry | bool | None


class BrokRPCError(Exception):
    pass


class BrokerError(BrokRPCError):
    pass


class PublisherError(BrokRPCError):
    pass


class ConsumerError(BrokRPCError):
    pass


class SerializerError(BrokRPCError):
    pass


class SerializerLoadError(SerializerError, ValueError):
    pass


class SerializerDumpError(SerializerError, ValueError):
    pass



from __future__ import annotations

import asyncio
import enum
import inspect
import typing as t
import warnings
from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass
from functools import partial
from signal import Signals

if t.TYPE_CHECKING:
    from concurrent.futures import Executor
    from datetime import timedelta

    from brokrpc.abc import BinaryConsumer, BinaryPublisher, BoundConsumer
    from brokrpc.broker import Broker

from brokrpc.options import BindingOptions, ExchangeOptions, QueueOptions, merge_options
from brokrpc.rpc.abc import HandlerSerializer, UnaryUnaryHandler
from brokrpc.rpc.handler import AsyncFuncHandler, SyncFuncHandler
from brokrpc.rpc.model import Request, ServerError
from brokrpc.stringify import to_str_obj


class State(enum.Enum):
    UNKNOWN = enum.auto()
    IDLE = enum.auto()
    STARTUP = enum.auto()
    RUNNING = enum.auto()
    CLEANUP = enum.auto()


class ServerNotInConfigurableStateError(ServerError):
    pass


class ServerStartupError(ServerError, BaseExceptionGroup):
    pass


@dataclass(frozen=True, kw_only=True)
class ServerOptions:
    startup_timeout: timedelta | None = None
    cleanup_timeout: timedelta | None = None
    executor: Executor | None = None


class Server:
    def __init__(
        self,
        broker: Broker,
        options: ServerOptions | None = None,
    ) -> None:
        self.__broker = broker
        self.__options = options

        self.__lock = asyncio.Lock()
        self.__cm_stack = AsyncExitStack()
        self.__state = State.IDLE
        self.__handlers: list[t.AsyncContextManager[BoundConsumer]] = []
        self.__bound_consumers: list[BoundConsumer] = []

    def __str__(self) -> str:
        return to_str_obj(self, state=self.__state.name, broker=self.__broker)

    def __del__(self) -> None:
        if self.__state is not State.IDLE:
            warnings.warn("server was not stopped properly", RuntimeWarning, stacklevel=1)

    @property
    def state(self) -> State:
        return self.__state

    def register_unary_unary_handler[U, V](
        self,
        *,
        func: UnaryUnaryHandler[Request[U], V] | t.Callable[[Request[U]], t.Awaitable[V]] | t.Callable[[Request[U]], V],
        routing_key: str,
        serializer: HandlerSerializer[U, V],
        exchange: ExchangeOptions | None = None,
        queue: QueueOptions | None = None,
    ) -> None:
        if self.__state is not State.IDLE:
            raise ServerNotInConfigurableStateError(self)

        factory: t.Callable[[BinaryPublisher], BinaryConsumer]
        match func:
            case handler if isinstance(handler, UnaryUnaryHandler):
                factory = partial(AsyncFuncHandler, handler.handle, serializer)

            case async_func if inspect.iscoroutinefunction(async_func):
                factory = partial(AsyncFuncHandler, async_func, serializer)

            case sync_func if callable(sync_func):
                factory = partial(
                    SyncFuncHandler,
                    t.cast(t.Callable[[Request[U]], V], sync_func),
                    serializer,
                    executor=self.__options.executor if self.__options is not None else None,
                )

            case _:
                # TODO: make `t.assert_never` work
                details = "invalid func type"
                raise TypeError(details, func)

        self.__handlers.append(
            self.__bind_consumer(
                factory=factory,
                binding=self.__get_binding_options(exchange, routing_key, queue),
            )
        )

    async def start(self) -> None:
        async with self.__lock:
            if self.__state is State.RUNNING:
                return

            self.__cm_stack.callback(self.__bound_consumers.clear)

            self.__state = State.STARTUP
            try:
                bound_consumers = await asyncio.gather(
                    *(self.__cm_stack.enter_async_context(handler_cm) for handler_cm in self.__handlers)
                )

            except Exception:
                self.__state = State.UNKNOWN
                raise

            else:
                self.__bound_consumers.extend(bound_consumers)
                self.__state = State.RUNNING

    async def stop(self) -> None:
        async with self.__lock:
            if self.__state not in {State.UNKNOWN, State.RUNNING}:
                return

            self.__state = State.CLEANUP
            try:
                await self.__cm_stack.aclose()

            except Exception:
                self.__state = State.UNKNOWN
                raise

            else:
                self.__state = State.IDLE

    @asynccontextmanager
    async def run(self) -> t.AsyncIterator[Server]:
        try:
            await self.start()
            yield self

        finally:
            await self.stop()

    async def run_until_done(self, done: asyncio.Event) -> None:
        async with self.run():
            await done.wait()

    async def run_until_terminated(
        self,
        signals: t.Collection[Signals] = (Signals.SIGTERM, Signals.SIGINT),
    ) -> None:
        done = asyncio.Event()

        loop = asyncio.get_running_loop()

        for sig in signals:
            loop.add_signal_handler(sig, done.set)

        try:
            await self.run_until_done(done)

        finally:
            for sig in signals:
                loop.remove_signal_handler(sig)

    @asynccontextmanager
    async def __bind_consumer(
        self,
        factory: t.Callable[[BinaryPublisher], BinaryConsumer],
        binding: BindingOptions,
    ) -> t.AsyncIterator[BoundConsumer]:
        async with (
            self.__broker.publisher() as replier,
            self.__broker.consumer(factory(replier), binding) as bound_consumer,
        ):
            yield bound_consumer

    def __get_binding_options(
        self,
        exchange: ExchangeOptions | None,
        routing_key: str,
        queue: QueueOptions | None,
    ) -> BindingOptions:
        return BindingOptions(
            exchange=exchange,
            binding_keys=(routing_key,),
            queue=merge_options(
                QueueOptions(
                    name=f"requests.{routing_key}",
                ),
                queue,
            ),
        )
